import json
import logging
import re
import regex
from abc import ABC
from typing import Dict, List

import aceutils.dict_and_list_util as dict_and_list_util
import aceutils.misc_utils as misc_utils
import aceutils.string_util as string_util
from acedeploy.core.model_configuration import ObjectOption, TableLikeObjectOption, TableObjectOption
from acedeploy.core.model_database_object import DatabaseObject
from acedeploy.core.model_sql_entities import DbObjectType, PolicyType
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class ColumnInstance(object):
    """
    Database object type Table-, DynamicTable-, or View- Column
    """

    def __init__(self, metadata_query_result: Dict, quote_identifiers: bool = False):
        self._quote_identifiers = quote_identifiers
        self.database_name = metadata_query_result["DATABASE_NAME"]
        self.object_schema = metadata_query_result["TABLE_SCHEMA"]
        self.object_name = metadata_query_result["TABLE_NAME"]
        self.column_name = metadata_query_result["COLUMN_NAME"]
        self.ordinal_position = metadata_query_result["ORDINAL_POSITION"]
        column_default_raw = metadata_query_result.get("COLUMN_DEFAULT")
        if (
            column_default_raw
            and column_default_raw.endswith(
                ".NEXTVAL"
            )  # if the default is a nextval, the result will contain the database name, which must be removed for comparisons to work
            and column_default_raw.startswith(
                f"{self.database_name}."
            )  # only remove database name if the sequence is in the same database
        ):
            self.column_default = string_util.remove_prefix(
                column_default_raw, f"{self.database_name}."
            )
        else:
            self.column_default = column_default_raw
        self.is_nullable = metadata_query_result.get("IS_NULLABLE") # IS_NULLABLE is not set for DynamicTable-Columns, because it is not supported in the DDL of a DynamicTable and is only inherited from underlying objects
        self.data_type = metadata_query_result["DATA_TYPE"]
        self.character_maximum_length = metadata_query_result.get(
            "CHARACTER_MAXIMUM_LENGTH"
        )
        self.character_octet_length = metadata_query_result.get(
            "CHARACTER_OCTET_LENGTH"
        )
        self.numeric_precision = metadata_query_result.get("NUMERIC_PRECISION")
        self.numeric_precision_radix = metadata_query_result.get(
            "NUMERIC_PRECISION_RADIX"
        )
        self.numeric_scale = metadata_query_result.get("NUMERIC_SCALE")
        self.datetime_precision = metadata_query_result.get("DATETIME_PRECISION")
        self.interval_type = metadata_query_result.get("INTERVAL_TYPE")
        self.interval_precision = metadata_query_result.get("INTERVAL_PRECISION")
        self.character_set_catalog = metadata_query_result.get("CHARACTER_SET_CATALOG")
        self.character_set_schema = metadata_query_result.get("CHARACTER_SET_SCHEMA")
        self.character_set_name = metadata_query_result.get("CHARACTER_SET_NAME")
        self.collation_catalog = metadata_query_result.get("COLLATION_CATALOG")
        self.collation_schema = metadata_query_result.get("COLLATION_SCHEMA")
        self.collation_name = metadata_query_result.get("COLLATION_NAME")
        self.domain_catalog = metadata_query_result.get("DOMAIN_CATALOG")
        self.domain_schema = metadata_query_result.get("DOMAIN_SCHEMA")
        self.domain_name = metadata_query_result.get("DOMAIN_NAME")
        self.udt_catalog = metadata_query_result.get("UDT_CATALOG")
        self.udt_schema = metadata_query_result.get("UDT_SCHEMA")
        self.udt_name = metadata_query_result.get("UDT_NAME")
        self.scope_catalog = metadata_query_result.get("SCOPE_CATALOG")
        self.scope_schema = metadata_query_result.get("SCOPE_SCHEMA")
        self.scope_name = metadata_query_result.get("SCOPE_NAME")
        self.maximum_cardinality = metadata_query_result.get("MAXIMUM_CARDINALITY")
        self.dtd_identifier = metadata_query_result.get("DTD_IDENTIFIER")
        self.is_self_referencing = metadata_query_result["IS_SELF_REFERENCING"]
        self.is_identity = metadata_query_result["IS_IDENTITY"]
        self.identity_generation = metadata_query_result.get("IDENTITY_GENERATION")
        self.identity_start = metadata_query_result.get("IDENTITY_START")
        self.identity_increment = metadata_query_result.get("IDENTITY_INCREMENT")
        self.identity_maximum = metadata_query_result.get("IDENTITY_MAXIMUM")
        self.identity_minimum = metadata_query_result.get("IDENTITY_MINIMUM")
        self.identity_cycle = metadata_query_result.get("IDENTITY_CYCLE")
        self.comment = metadata_query_result.get("COMMENT")
        self.tags = metadata_query_result.get("tags", {})

    @property
    def column_name_quoted(self):
        if self._quote_identifiers:
            return f'"{self.column_name}"'
        return self.column_name

    def __eq__(self, other):
        """
        Test if two Column objects are identical.
        Tests for all properties, except ordinal_position and database_name.
        """
        return (
            self.object_schema == other.object_schema
            and self.object_name == other.object_name
            and self.column_name == other.column_name
            and self.column_default == other.column_default
            and self.is_nullable == other.is_nullable
            and self.data_type == other.data_type
            and self.character_maximum_length == other.character_maximum_length
            and self.character_octet_length == other.character_octet_length
            and self.numeric_precision == other.numeric_precision
            and self.numeric_precision_radix == other.numeric_precision_radix
            and self.numeric_scale == other.numeric_scale
            and self.datetime_precision == other.datetime_precision
            and self.interval_type == other.interval_type
            and self.interval_precision == other.interval_precision
            and self.character_set_catalog == other.character_set_catalog
            and self.character_set_schema == other.character_set_schema
            and self.character_set_name == other.character_set_name
            and self.collation_catalog == other.collation_catalog
            and self.collation_schema == other.collation_schema
            and self.collation_name == other.collation_name
            and self.domain_catalog == other.domain_catalog
            and self.domain_schema == other.domain_schema
            and self.domain_name == other.domain_name
            and self.udt_catalog == other.udt_catalog
            and self.udt_schema == other.udt_schema
            and self.udt_name == other.udt_name
            and self.scope_catalog == other.scope_catalog
            and self.scope_schema == other.scope_schema
            and self.scope_name == other.scope_name
            and self.maximum_cardinality == other.maximum_cardinality
            and self.dtd_identifier == other.dtd_identifier
            and self.is_self_referencing == other.is_self_referencing
            and self.is_identity == other.is_identity
            and self.identity_generation == other.identity_generation
            and self.identity_start == other.identity_start
            and self.identity_increment == other.identity_increment
            and self.identity_maximum == other.identity_maximum
            and self.identity_minimum == other.identity_minimum
            and self.identity_cycle == other.identity_cycle
            and self.tags == other.tags
            and (self.comment if self.comment else "")
            == (
                other.comment if other.comment else ""
            )  # comment can be None, treat as empty string for comparison
        )

    @staticmethod
    def get_differences(column1, column2):
        column1_dict = column1.__dict__.copy()
        column2_dict = column2.__dict__.copy()
        column1_dict["comment"] = (
            column1_dict["comment"] if column1_dict["comment"] else ""
        )  # comment can be None, treat as empty string for comparison
        column2_dict["comment"] = (
            column2_dict["comment"] if column2_dict["comment"] else ""
        )
        only_in_column1, only_in_column2 = dict_and_list_util.compare_nested_dicts(
            column1_dict, column2_dict
        )
        only_in_column1.pop("ordinal_position", 0)
        only_in_column1.pop("database_name", "")
        only_in_column2.pop("ordinal_position", 0)
        only_in_column2.pop("database_name", "")
        return only_in_column1, only_in_column2


class ConstraintColumn(object):
    def __init__(self, key_sequence: int, column_name: str, quote_identifiers: bool = False) -> None:
        self._quote_identifiers = quote_identifiers
        self.key_sequence = key_sequence
        self.column_name = column_name

    @property
    def column_name_quoted(self):
        if self._quote_identifiers:
            return f'"{self.column_name}"'
        return self.column_name

    def __eq__(self, other: "ConstraintColumn") -> bool:
        return (
            self.key_sequence == other.key_sequence
            and self.column_name == other.column_name
        )


class ConstraintColumnForeignKey(ConstraintColumn):
    def __init__(
        self, key_sequence: int, column_name: str, pk_column_name: str, quote_identifiers: bool = False
    ) -> None:
        super().__init__(key_sequence, column_name, quote_identifiers)
        self.pk_column_name = pk_column_name

    @property
    def pk_column_name_quoted(self):
        if self._quote_identifiers:
            return f'"{self.pk_column_name}"'
        return self.pk_column_name

    def __eq__(self, other: "ConstraintColumnForeignKey") -> bool:
        return super().__eq__(other) and self.pk_column_name == other.pk_column_name


class InstanceConstraint(ABC):
    def __init__(self, metadata: List[Dict], quote_column_identifiers: bool = False) -> None:
        self._quote_column_identifiers = quote_column_identifiers
        if len(metadata) == 0:
            raise ValueError("Cannot create object without metadata")
        self.schema_name = metadata[0]["schema_name"]
        if any(m["schema_name"] != self.schema_name for m in metadata):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table"
            )
        self.table_name = metadata[0]["table_name"]
        if any(m["table_name"] != self.table_name for m in metadata):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table"
            )
        self.constraint_name = metadata[0]["constraint_name"]
        if any(m["constraint_name"] != self.constraint_name for m in metadata):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table"
            )
        self.comment = metadata[0]["comment"]
        if any(m["comment"] != self.comment for m in metadata):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table"
            )
        self.columns = self._get_columns(metadata)

    def _get_columns(self, metadata: List[Dict]) -> List[ConstraintColumn]:
        return [ConstraintColumn(m["key_sequence"], m["column_name"], self._quote_column_identifiers) for m in metadata]

    @staticmethod
    def factory(
        metadata_query_result: List[Dict], constraint_type: str, quote_column_identifiers: bool,
    ) -> "InstanceConstraint":
        mapping = {
            "foreign_key": {
                "class": InstanceConstraintForeignKey,
                "constraint_name_key": "fk_name",
            },
            "primary_key": {
                "class": InstanceConstraintPrimaryKey,
                "constraint_name_key": "constraint_name",
            },
            "unique_key": {
                "class": InstanceConstraintUniqueKey,
                "constraint_name_key": "constraint_name",
            },
        }
        constraint_objects = []
        constraint_names = set(
            [
                m[mapping[constraint_type]["constraint_name_key"]]
                for m in metadata_query_result
            ]
        )
        for constraint_name in constraint_names:
            metadata = [
                m
                for m in metadata_query_result
                if m[mapping[constraint_type]["constraint_name_key"]] == constraint_name
            ]
            constraint_objects.append(mapping[constraint_type]["class"](metadata, quote_column_identifiers=quote_column_identifiers))
        return constraint_objects

    def __eq__(self, other: "InstanceConstraint") -> bool:
        self_name = (
            "system_assigned"
            if self._is_system_assigned_name(self.constraint_name)
            else self.constraint_name
        )
        other_name = (
            "system_assigned"
            if self._is_system_assigned_name(other.constraint_name)
            else other.constraint_name
        )
        return (
            self.schema_name == other.schema_name
            and self.table_name == other.table_name
            and self_name == other_name
            and self.comment == other.comment
            and self.columns == other.columns
        )

    @staticmethod
    def _is_system_assigned_name(name: str) -> str:
        """
        Test if the given string is a system assigned constraint name.
        """
        if re.match(
            r"^SYS_CONSTRAINT_[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$",
            name,
        ):
            return True
        else:
            return False


class InstanceConstraintPrimaryKey(InstanceConstraint):
    pass


class InstanceConstraintUniqueKey(InstanceConstraint):
    pass


class InstanceConstraintForeignKey(InstanceConstraint):
    def __init__(self, metadata: List[Dict], quote_column_identifiers: bool = False) -> None:
        self._quote_column_identifiers = quote_column_identifiers
        metadata_copy = [m.copy() for m in metadata]
        # need to rename metadata entries to match expected key names in InstanceConstraint
        for m in metadata_copy:
            m["schema_name"] = m.pop("fk_schema_name")
            m["table_name"] = m.pop("fk_table_name")
            m["constraint_name"] = m.pop("fk_name")
        super().__init__(metadata_copy, quote_column_identifiers)
        self.pk_schema_name = metadata_copy[0]["pk_schema_name"]
        if any(m["pk_schema_name"] != self.pk_schema_name for m in metadata_copy):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table referencing a single pk"
            )
        self.pk_table_name = metadata_copy[0]["pk_table_name"]
        if any(m["pk_table_name"] != self.pk_table_name for m in metadata_copy):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table referencing a single pk"
            )
        self.pk_name = metadata_copy[0]["pk_name"]
        if any(m["pk_name"] != self.pk_name for m in metadata_copy):
            raise ValueError(
                "Metadata must refer to a single constraint on a single table referencing a single pk"
            )

    @property
    def fk_schema_name(self):
        return self.schema_name

    @property
    def fk_table_name(self):
        return self.table_name

    @property
    def fk_name(self):
        return self.constraint_name

    def _get_columns(self, metadata):
        return [
            ConstraintColumnForeignKey(
                m["key_sequence"], m["fk_column_name"], m["pk_column_name"],
                self._quote_column_identifiers
            )
            for m in metadata
        ]

    def __eq__(self, other: "InstanceConstraintForeignKey") -> bool:
        self_pk_name = (
            "system_assigned"
            if self._is_system_assigned_name(self.pk_name)
            else self.pk_name
        )
        other_pk_name = (
            "system_assigned"
            if self._is_system_assigned_name(other.pk_name)
            else other.pk_name
        )
        return (
            super().__eq__(other)
            and self.pk_schema_name == other.pk_schema_name
            and self.pk_table_name == other.pk_table_name
            and self_pk_name == other_pk_name
        )


class InstanceObject(DatabaseObject):
    """
    Base object for storing basic information about a database object
    Only for inherit usage
    """

    def __init__(self, result, object_type: DbObjectType, object_options: ObjectOption=None):
        """
            inits a new instance
        Args:
            result: any - resultset entry for this entity
        """
        super().__init__(result["SCHEMA_NAME"], result["OBJECT_NAME"], object_type)
        self.database_name = result["DATABASE_NAME"].upper()
        self.object_options = object_options

    def __eq__(self, other: "InstanceObject") -> bool:
        return (
            self.object_type == other.object_type
            and self.schema == other.schema
            and self.name == other.name
        )

    @staticmethod
    def factory(
        object_type: DbObjectType,
        metadata_query_result: Dict,
        object_options: ObjectOption=None,
        **kwargs
    ):
        mapping = {
            DbObjectType.TABLE: InstanceTable,
            DbObjectType.EXTERNALTABLE: InstanceExternalTable,
            DbObjectType.VIEW: InstanceView,
            DbObjectType.MATERIALIZEDVIEW: InstanceMaterializedView,
            DbObjectType.FUNCTION: InstanceFunction,
            DbObjectType.PROCEDURE: InstanceProcedure,
            DbObjectType.STAGE: InstanceStage,
            DbObjectType.FILEFORMAT: InstanceFileformat,
            DbObjectType.SCHEMA: InstanceSchema,
            DbObjectType.STREAM: InstanceStream,
            DbObjectType.TASK: InstanceTask,
            DbObjectType.PIPE: InstancePipe,
            DbObjectType.SEQUENCE: InstanceSequence,
            DbObjectType.MASKINGPOLICY: InstanceMaskingPolicy,
            DbObjectType.ROWACCESSPOLICY: InstanceRowAccessPolicy,
            DbObjectType.DYNAMICTABLE: InstanceDynamicTable,
            DbObjectType.HYBRIDTABLE: InstanceHybridTable,
            DbObjectType.NETWORKRULE: InstanceNetworkRule,
            DbObjectType.TAG: InstanceTag,
        }
        return mapping[object_type](
            metadata_query_result=metadata_query_result,
            object_options=object_options,
            **kwargs,
        )


class InstanceTableLike(InstanceObject):
    """
    Database object type Table or View
    """

    def __init__(
        self,
        metadata_query_result: Dict,
        object_type: DbObjectType,
        object_options: TableLikeObjectOption=None,
        ignore_column_order: bool = False,
    ):
        super().__init__(metadata_query_result, object_type, object_options)
        self._quote_column_identifiers = object_options.quoteColumnIdentifiers if object_options and hasattr(object_options, 'quoteColumnIdentifiers') else False
        self.table_type = metadata_query_result["TABLE_TYPE"]
        clustering_key_raw = metadata_query_result["CLUSTERING_KEY"]
        self.clustering_key = (
            clustering_key_raw.upper() if clustering_key_raw else clustering_key_raw
        )
        self.ignore_column_order = ignore_column_order
        self.table_columns = sorted(
            [ColumnInstance(column, self._quote_column_identifiers) for column in metadata_query_result["COLUMN_DETAILS"]],
            key=lambda c: c.ordinal_position,
        )

    def __eq__(self, other: "InstanceTableLike") -> bool:
        """
        Test if two Table objects are identical.
        Tests for all properties, except database_name.
        """
        clustering_key_self = (
            self.clustering_key.replace(" ", "").replace("\n", "").replace("\t", "")
            if isinstance(self.clustering_key, str)
            else self.clustering_key
        )  # clustering key can be, e.g. LINEAR(a,b) or LINEAR(a, b)
        clustering_key_other = (
            other.clustering_key.replace(" ", "").replace("\n", "").replace("\t", "")
            if isinstance(other.clustering_key, str)
            else other.clustering_key
        )
        if (
            (not super().__eq__(other))
            or self.table_type != other.table_type
            or clustering_key_self != clustering_key_other
        ):
            return False

        if len(self.table_columns) != len(other.table_columns):
            return False

        if self.ignore_column_order:
            if not InstanceTableLike._compare_columns_ignore_order(
                self.table_columns, other.table_columns
            ):
                return False
        else:
            if not InstanceTableLike._compare_columns_with_order(
                self.table_columns, other.table_columns
            ):
                return False

        return True

    @staticmethod
    def _compare_columns_with_order(columns_self, columns_other):
        for column_self, column_other in zip(columns_self, columns_other):
            if column_self != column_other:
                return False
        return True

    @staticmethod
    def _compare_columns_ignore_order(columns_self, columns_other):
        comparison = InstanceTableLike._compare_columns(columns_self, columns_other)
        return comparison == {
            "columns_not_in_both": {"only_in_table1": [], "only_in_table2": []},
            "property_differences": {"table1": {}, "table2": {}},
        }

    @staticmethod
    def generate_diff_description(table1, table2):
        nested_result_dict = InstanceTableLike._get_diffs(table1, table2)
        return json.dumps(
            dict_and_list_util.strip_nested_dict(nested_result_dict), indent=4
        )

    @staticmethod
    def _get_diffs(table1, table2):
        table1_dict = table1.__dict__.copy()
        table1_dict.pop("database_name")
        table1_dict.pop("table_columns")
        table1_dict.pop("ignore_column_order")
        table2_dict = table2.__dict__.copy()
        table2_dict.pop("database_name")
        table2_dict.pop("table_columns")
        table2_dict.pop("ignore_column_order")

        x = table1_dict.pop(
            "constraints_foreign_key", None
        )  # TODO: implement getting diff on constraints
        table1_dict.pop("constraints_primary_key", None)
        table1_dict.pop("constraints_unique_key", None)
        table2_dict.pop("constraints_foreign_key", None)
        table2_dict.pop("constraints_primary_key", None)
        table2_dict.pop("constraints_unique_key", None)

        only_in_table1, only_in_table2 = dict_and_list_util.compare_nested_dicts(
            table1_dict, table2_dict
        )
        column_differences = InstanceTableLike._compare_columns(
            table1.table_columns, table2.table_columns
        )

        result = {}
        result["property_differences"] = {
            "table1": only_in_table1,
            "table2": only_in_table2,
        }
        result["column_differences"] = column_differences
        if x is not None:  # TODO: implement getting diff on constraints
            result[
                "constraint_differences"
            ] = "currently not implemented, check manually"
        return result

    @staticmethod
    def _compare_columns(column_list1, column_list2):

        columns_only_in_1 = [
            c.column_name
            for c in column_list1
            if c.column_name not in [cn.column_name for cn in column_list2]
        ]
        columns_only_in_2 = [
            c.column_name
            for c in column_list2
            if c.column_name not in [cn.column_name for cn in column_list1]
        ]

        result = {}
        result["columns_not_in_both"] = {
            "only_in_table1": columns_only_in_1,
            "only_in_table2": columns_only_in_2,
        }
        result["property_differences"] = {"table1": {}, "table2": {}}

        column_names_in_both = [
            c.column_name
            for c in column_list1
            if c.column_name in [cn.column_name for cn in column_list2]
        ]
        for column_name in column_names_in_both:
            column1 = next(
                c
                for c in column_list1
                if c.column_name == column_name
            )
            column2 = next(
                c
                for c in column_list2
                if c.column_name == column_name
            )
            only_in_column1, only_in_column2 = ColumnInstance.get_differences(
                column1, column2
            )
            if len(only_in_column1) > 0:
                result["property_differences"]["table1"][column_name] = only_in_column1
            if len(only_in_column2) > 0:
                result["property_differences"]["table2"][column_name] = only_in_column2

        return result


class InstanceTable(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: TableObjectOption=None, ignore_column_order: bool = False):
        super().__init__(metadata_query_result, DbObjectType.TABLE, object_options, ignore_column_order)
        self.row_count = metadata_query_result["ROW_COUNT"] # should not be used for comparsions
        self.bytes = metadata_query_result["BYTES"] # should not be used for comparsions
        self.retention_time = metadata_query_result["RETENTION_TIME"]
        self.schema_retention_time = metadata_query_result["SCHEMA_RETENTION_TIME"]
        self.comment = metadata_query_result["COMMENT"]
        self.constraints_foreign_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_foreign_keys"],
            constraint_type="foreign_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.constraints_primary_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_primary_keys"],
            constraint_type="primary_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.constraints_unique_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_unique_keys"],
            constraint_type="unique_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.tags = metadata_query_result.get("tags", {})

        pr_factory = PolicyReferenceFactory()
        policy_references = [
                pr_factory.factory(ref) for ref in metadata_query_result.get("policy_references", [])
            ]
        self.row_access_policy_references = [
            ref for ref in policy_references if ref.policy_type == PolicyType.ROWACCESS
        ]
        self.masking_policy_references = [
            ref for ref in policy_references if ref.policy_type == PolicyType.MASKING
        ]

    def __str__(self):
        return f"InstanceTable: {self.id}"

    def __repr__(self):
        return f"InstanceTable: {self.id}"

    def __eq__(self, other: "InstanceTable") -> bool:
        if not (
            super().__eq__(other)
            and self.retention_time == other.retention_time
            and self.schema_retention_time == other.schema_retention_time
            and self.comment == other.comment
            and self.constraints_foreign_key == other.constraints_foreign_key
            and self.constraints_primary_key == other.constraints_primary_key
            and self.constraints_unique_key == other.constraints_unique_key
            and self.tags == other.tags
        ):
            return False
    
        if set(self.row_access_policy_references) != set(other.row_access_policy_references):
            return False
        
        if set(self.masking_policy_references) != set(other.masking_policy_references):
            return False

        return True


class InstanceHybridTable(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: TableObjectOption=None, ignore_column_order: bool = False):
        super().__init__(metadata_query_result, DbObjectType.HYBRIDTABLE, object_options, ignore_column_order)
        self.row_count = metadata_query_result["ROW_COUNT"] # should not be used for comparsions
        self.bytes = metadata_query_result["BYTES"] # should not be used for comparsions
        self.retention_time = metadata_query_result["RETENTION_TIME"]
        self.schema_retention_time = metadata_query_result["SCHEMA_RETENTION_TIME"]
        self.comment = metadata_query_result["COMMENT"]
        self.constraints_foreign_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_foreign_keys"],
            constraint_type="foreign_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.constraints_primary_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_primary_keys"],
            constraint_type="primary_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.constraints_unique_key = InstanceConstraint.factory(
            metadata_query_result=metadata_query_result["constraint_unique_keys"],
            constraint_type="unique_key",
            quote_column_identifiers=self._quote_column_identifiers,
        )
        self.tags = metadata_query_result.get("tags", {})

        pr_factory = PolicyReferenceFactory()
        policy_references = [
                pr_factory.factory(ref) for ref in metadata_query_result.get("policy_references", [])
            ]
        self.row_access_policy_references = [
            ref for ref in policy_references if ref.policy_type == PolicyType.ROWACCESS
        ]
        self.masking_policy_references = [
            ref for ref in policy_references if ref.policy_type == PolicyType.MASKING
        ]

    def __str__(self):
        return f"InstanceHybridTable: {self.id}"

    def __repr__(self):
        return f"InstanceHybridTable: {self.id}"

    def __eq__(self, other: "InstanceHybridTable") -> bool:
        if not (
            super().__eq__(other)
            and self.retention_time == other.retention_time
            and self.schema_retention_time == other.schema_retention_time
            and self.comment == other.comment
            and self.constraints_foreign_key == other.constraints_foreign_key
            and self.constraints_primary_key == other.constraints_primary_key
            and self.constraints_unique_key == other.constraints_unique_key
            and self.tags == other.tags
        ):
            return False
    
        if set(self.row_access_policy_references) != set(other.row_access_policy_references):
            return False
        
        if set(self.masking_policy_references) != set(other.masking_policy_references):
            return False

        return True


class InstanceView(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: TableLikeObjectOption=None, ignore_column_order: bool = False):
        super().__init__(metadata_query_result, DbObjectType.VIEW, object_options, ignore_column_order)
        self.view_definition = metadata_query_result["VIEW_DEFINITION"]
        self.is_secure = metadata_query_result["IS_SECURE"]
        self.comment = metadata_query_result["COMMENT"]
        applied_policies_metadata_legacy = metadata_query_result.get(
            "APPLIED_POLICIES_LEGACY", []
        )  # applied policies property is optional
        self.applied_policies_legacy = [
            AppliedPolicyLegacy.factory(apm) for apm in applied_policies_metadata_legacy
        ]
        self.tags = metadata_query_result.get("tags", {})

    def __str__(self):
        return f"InstanceView: {self.id}"

    def __repr__(self):
        return f"InstanceView: {self.id}"

    def __eq__(self, other: "InstanceView") -> bool:
        return (
            super().__eq__(other)
            and string_util.compare_strings_ignore_whitespace_and_case(
                string_util.add_copy_grants(string_util.add_create_or_replace(self.view_definition)),
                string_util.add_copy_grants(string_util.add_create_or_replace(other.view_definition))
                )
            # and self.comment == other.comment #TODO -> comments on views are currently (10.04.2024) ignored by the MetadataService (view_definition in the information_schema.views does not include comments on views)
            and self.is_secure == other.is_secure
            and self.tags == other.tags
        )
    

class InstanceMaterializedView(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None, ignore_column_order: bool = False,):
        super().__init__(
            metadata_query_result, DbObjectType.MATERIALIZEDVIEW, object_options, ignore_column_order
        )
        self.comment = metadata_query_result["COMMENT"]
        self.tags = metadata_query_result.get("tags", {})

    def __str__(self):
        return f"InstanceMaterializedView: {self.id}"

    def __repr__(self):
        return f"InstanceMaterializedView: {self.id}"

    def __eq__(self, other: "InstanceMaterializedView") -> bool:
        return (
            super().__eq__(other) 
            and self.comment == other.comment 
            and self.tags == other.tags
        )


class InstanceExternalTable(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None, ignore_column_order: bool = False):
        super().__init__(
            metadata_query_result, DbObjectType.EXTERNALTABLE, object_options, ignore_column_order
        )
        self.comment = metadata_query_result["COMMENT"]
        self.location = metadata_query_result["LOCATION"]
        self.file_format_name = metadata_query_result["FILE_FORMAT_NAME"]
        self.file_format_type = metadata_query_result["FILE_FORMAT_TYPE"]
        self.tags = metadata_query_result.get("tags", {})

    def __str__(self):
        return f"InstanceExternalTable: {self.id}"

    def __repr__(self):
        return f"InstanceExternalTable: {self.id}"

    def __eq__(self, other: "InstanceExternalTable") -> bool:
        return (
            super().__eq__(other)
            and self.comment == other.comment
            and self.location == other.location
            and self.file_format_name == other.file_format_name
            and self.file_format_type == other.file_format_type
            and self.tags == other.tags
        )


class InstanceStage(InstanceObject):
    """
    Database object typ Stage
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        metadata_query_result["DATABASE_NAME"] = metadata_query_result["database_name"]
        metadata_query_result["SCHEMA_NAME"] = metadata_query_result["schema_name"]
        metadata_query_result["OBJECT_NAME"] = metadata_query_result["name"]
        super().__init__(metadata_query_result, DbObjectType.STAGE, object_options)
        # self.url = metadata_query_result['url'] # also appears in dict as STAGE_LOCATION
        self.has_credentials = metadata_query_result["has_credentials"]
        self.has_encryption_key = metadata_query_result["has_encryption_key"]
        self.comment = metadata_query_result["comment"]
        # self.region = metadata_query_result['region'] # do not use this attribute. snowflake sometimes does not return a value for this. region does not add value anyways, because it is part of stage_location anyways.
        self.type = metadata_query_result["type"]
        self.cloud = metadata_query_result["cloud"]
        # self.storage_integration = metadata_query_result['storage_integration'] # also appears in dict as STAGE_INTEGRATION
        self.stage_file_format = metadata_query_result["STAGE_FILE_FORMAT"]  # dict
        self.stage_copy_options = metadata_query_result["STAGE_COPY_OPTIONS"]  # dict
        self.stage_location = metadata_query_result["STAGE_LOCATION"]  # dict
        self.stage_integration = metadata_query_result["STAGE_INTEGRATION"]  # dict
        self.directory = metadata_query_result["DIRECTORY"]  # dict
        self.directory.pop("LAST_REFRESHED_ON", None) # not relevant for metadata comparison

    def __str__(self):
        return f"InstanceStage: {self.id}"

    def __repr__(self):
        return f"InstanceStage: {self.id}"

    def __eq__(self, other: "InstanceStage") -> bool:
        return (
            super().__eq__(other)
            # and self.url == other.url
            and self.has_credentials == other.has_credentials
            and self.has_encryption_key == other.has_encryption_key
            and self.comment == other.comment
            # and self.region == other.region # do not compare this attribute. snowflake sometimes does not return a value for this. region does not add value anyways, because it is part of stage_location anyways.
            and self.type == other.type
            and self.cloud == other.cloud
            # and self.storage_integration == other.storage_integration
            and self.stage_file_format == other.stage_file_format
            and self.stage_copy_options == other.stage_copy_options
            and self.stage_location == other.stage_location
            and self.stage_integration == other.stage_integration
            and self.directory == other.directory
        )


class InstanceFileformat(InstanceObject):
    """
    Database object typ File format
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        metadata_query_result["DATABASE_NAME"] = metadata_query_result["database_name"]
        metadata_query_result["SCHEMA_NAME"] = metadata_query_result["schema_name"]
        metadata_query_result["OBJECT_NAME"] = metadata_query_result["name"]
        super().__init__(metadata_query_result, DbObjectType.FILEFORMAT, object_options)
        self.file_format_type = metadata_query_result["type"]
        self.format_options = self._convert_options_json_to_dict(
            metadata_query_result["format_options"]
        )
        self.comment = metadata_query_result["comment"]

    def __str__(self):
        return f"InstanceFileformat: {self.id}"

    def __repr__(self):
        return f"InstanceFileformat: {self.id}"

    def __eq__(self, other: "InstanceFileformat") -> bool:
        return (
            super().__eq__(other)
            and self.file_format_type == other.file_format_type
            and self.format_options == other.format_options
            and self.comment == other.comment
        )

    @staticmethod
    def _convert_options_json_to_dict(s: str) -> dict:
        d = json.loads(s)
        d.pop("TYPE")
        return d


class InstanceStream(InstanceObject):
    """
    Database object type Stream
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        metadata_query_result["DATABASE_NAME"] = metadata_query_result["database_name"]
        metadata_query_result["SCHEMA_NAME"] = metadata_query_result["schema_name"]
        metadata_query_result["OBJECT_NAME"] = metadata_query_result["name"]
        super().__init__(metadata_query_result, DbObjectType.STREAM, object_options)
        self.comment = metadata_query_result["comment"]
        self.table_name = self._remove_database_from_table_name(
            metadata_query_result["table_name"]
        )
        self.type = metadata_query_result["type"]
        self.stale = metadata_query_result["stale"]
        self.mode = metadata_query_result["mode"]

    def __eq__(self, other: "InstanceStream") -> bool:
        return (
            super().__eq__(other)
            and self.comment == other.comment
            and self.table_name == other.table_name
            and self.type == other.type
            and self.mode == other.mode
        )

    def __str__(self):
        return f"InstanceStream: {self.id}"

    def __repr__(self):
        return f"InstanceStream: {self.id}"

    @staticmethod
    def _remove_database_from_table_name(table_name):
        m = re.match(r"^\"?\w+\"?\.(?P<name>\"?\w+\"?\.\"?\w+\"?)$", table_name)
        if m:
            return m.group("name")
        else:
            return table_name


class InstanceTask(InstanceObject):
    """
    Database object typ File format
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        metadata_query_result["DATABASE_NAME"] = metadata_query_result["database_name"]
        metadata_query_result["SCHEMA_NAME"] = metadata_query_result["schema_name"]
        metadata_query_result["OBJECT_NAME"] = metadata_query_result["name"]
        super().__init__(metadata_query_result, DbObjectType.TASK, object_options)
        self.comment = metadata_query_result["comment"]
        self.warehouse = metadata_query_result["warehouse"]
        self.schedule = metadata_query_result["schedule"]
        self.predecessors = self._parse_predecessors(
            metadata_query_result["predecessors"], self.database_name
        )
        self.state = metadata_query_result["state"]
        self.definition = metadata_query_result["definition"]
        self.condition = metadata_query_result["condition"]
        self.allow_overlapping_execution = metadata_query_result[
            "allow_overlapping_execution"
        ]

    def __str__(self):
        return f"InstanceTask: {self.id}"

    def __repr__(self):
        return f"InstanceTask: {self.id}"

    def __eq__(self, other: "InstanceTask") -> bool:
        return (
            super().__eq__(other)
            and self.comment == other.comment
            and self.warehouse == other.warehouse
            and self.schedule == other.schedule
            and self.predecessors == other.predecessors
            and self.state == other.state
            and self.definition == other.definition
            and self.condition == other.condition
            and self.allow_overlapping_execution == other.allow_overlapping_execution
        )

    @staticmethod
    def _parse_predecessors(metadata: str, database_name) -> List[str]:
        """
        Parse the metadata for predecessors.

        This function works both for accounts where behavior change 2022_03 is active and not active.
        https://community.snowflake.com/s/article/SHOW-TASKS-Command-and-TASK-DEPENDENTS-Function-PREDECESSORS-PREDECESSOR-Column
        """
        if metadata is None:  # pre bundle 2022_03 (no predecessor)
            return []
        elif metadata.startswith("["):  # post bundle 2022_03
            predecessor_list = json.loads(metadata)
            return [
                string_util.remove_prefix(p, f"{database_name}.")
                for p in predecessor_list
            ]
        else:  # pre bundle 2022_03 (with predecessor)
            return [string_util.remove_prefix(metadata, f"{database_name}.")]


class InstancePipe(InstanceObject):
    """
    Database object typ File format
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.PIPE, object_options)
        self.definition = metadata_query_result["DEFINITION"]
        self.is_autoingest_enabled = metadata_query_result["IS_AUTOINGEST_ENABLED"]
        self.notification_channel_name = metadata_query_result[
            "NOTIFICATION_CHANNEL_NAME"
        ]
        self.comment = metadata_query_result["COMMENT"]
        self.pattern = metadata_query_result["PATTERN"]
        self.integration = metadata_query_result["integration"]
        self.execution_state = metadata_query_result[
            "execution_state"
        ]  # RUNNING or PAUSED (or some error). Is set outside of deployment and does not need to be compared in __eq__()

    def __str__(self):
        return f"InstancePipe: {self.id}"

    def __repr__(self):
        return f"InstancePipe: {self.id}"

    def __eq__(self, other: "InstancePipe") -> bool:
        return (
            super().__eq__(other)
            and self.definition == other.definition
            and self.is_autoingest_enabled == other.is_autoingest_enabled
            and self.notification_channel_name == other.notification_channel_name
            and self.comment == other.comment
            and self.pattern == other.pattern
            and self.integration == other.integration
        )


class InstanceSequence(InstanceObject):
    """
    Database object type Sequence
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.SEQUENCE, object_options)
        self.data_type = metadata_query_result["DATA_TYPE"]
        self.numeric_precision = metadata_query_result["NUMERIC_PRECISION"]
        self.numeric_precision_radix = metadata_query_result["NUMERIC_PRECISION_RADIX"]
        self.numeric_scale = metadata_query_result["NUMERIC_SCALE"]
        self.start_value = metadata_query_result["START_VALUE"]
        self.increment = metadata_query_result["INCREMENT"]
        self.comment = metadata_query_result["COMMENT"]

    def __eq__(self, other: "InstanceSequence") -> bool:
        return (
            super().__eq__(other)
            and self.data_type == other.data_type
            and self.numeric_precision == other.numeric_precision
            and self.numeric_precision_radix == other.numeric_precision_radix
            and self.numeric_scale == other.numeric_scale
            and self.start_value == other.start_value
            and self.increment == other.increment
            and (self.comment if self.comment else "")
            == (
                other.comment if other.comment else ""
            )  # comment can be None, treat as empty string for comparison
        )

    def __str__(self):
        return f"InstanceSequence: {self.id}"

    def __repr__(self):
        return f"InstanceSequence: {self.id}"


class InstanceParameterObject(InstanceObject):
    """
    Database object type Function or Procedure
    """

    def __init__(self, metadata_query_result: Dict, object_type: DbObjectType, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, object_type, object_options)
        parameters = self.parse_parameters(metadata_query_result["SIGNATURE"])
        self.parameters = [
            misc_utils.map_datatype_name_to_default(p) for p in parameters
        ]
        self.data_type = metadata_query_result["DATA_TYPE"]
        self.character_maximum_length = metadata_query_result[
            "CHARACTER_MAXIMUM_LENGTH"
        ]
        self.character_octet_length = metadata_query_result["CHARACTER_OCTET_LENGTH"]
        self.numeric_precision = metadata_query_result["NUMERIC_PRECISION"]
        self.numeric_precision_radix = metadata_query_result["NUMERIC_PRECISION_RADIX"]
        self.numeric_scale = metadata_query_result["NUMERIC_SCALE"]
        self.language = metadata_query_result["LANGUAGE"]
        self.definition = metadata_query_result["DEFINITION"]

    def __eq__(self, other: "InstanceParameterObject") -> bool:
        return (
            super().__eq__(other)
            and self.parameters == other.parameters
            and self.data_type == other.data_type
            and self.character_maximum_length == other.character_maximum_length
            and self.character_octet_length == other.character_octet_length
            and self.numeric_precision == other.numeric_precision
            and self.numeric_precision_radix == other.numeric_precision_radix
            and self.numeric_scale == other.numeric_scale
            and self.language == other.language
            and self.definition == other.definition
        )

    @property
    def id(self):
        return f"{super().id} {self.parameters_string.upper()}"

    @property
    def full_name(self):
        return f"{super().full_name} {self.parameters_string.upper()}"

    @property
    def parameters_string(self):
        return f"({', '.join([s for s in self.parameters])})"

    @staticmethod
    def parse_parameters(signature: str) -> List[str]:
        """
            parses the snowflake parameter signature into valid tokens for snowflake connector
        Args:
            signature : str - snowflake parameter signature from query result set
        Raises:
            ValueError - if signature can not be parsed
        Returns:
            Data types from signature
        """
        if signature is None:
            raise ValueError("Signature can not be of type None.")

        output = []
        __argument_list = signature.replace("(", "").replace(")", "")
        if len(__argument_list) > 0:
            tokens = []
            if __argument_list.find(",") != -1:
                split_tokens = __argument_list.split(", ")
                for split_token in split_tokens:
                    tokens.append(split_token)
            else:
                tokens.append(__argument_list)

            for token in tokens:
                token_split = token.split(" ")
                if len(token_split) > 1:
                    output.append(token_split[1])
                else:
                    raise ValueError("Parameter token list not valid.")
        return output


class InstanceFunction(InstanceParameterObject):
    """
    Database object type Function
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.FUNCTION, object_options)
        self.is_external = metadata_query_result["IS_EXTERNAL"]
        self.is_secure = metadata_query_result["IS_SECURE"]
        self.volatility = metadata_query_result["VOLATILITY"]
        self.is_null_call = metadata_query_result["IS_NULL_CALL"]
        self.comment = metadata_query_result["COMMENT"]
        self.is_external = metadata_query_result["IS_EXTERNAL"]
        self.api_integration = metadata_query_result["API_INTEGRATION"]
        self.context_headers = metadata_query_result["CONTEXT_HEADERS"]
        self.max_batch_rows = metadata_query_result["MAX_BATCH_ROWS"]
        self.compression = metadata_query_result["COMPRESSION"]
        # the metadata that we currently query does not include information on Java functions, e.g. HANDLER
        # therefore, we can currently not use this class alone to determine if function instances are identical

    def __str__(self):
        return f"InstanceFunction: {self.id}"

    def __repr__(self):
        return f"InstanceFunction: {self.id}"

    def __eq__(self, other: "InstanceFunction") -> bool:
        return (
            super().__eq__(other)
            and self.is_external == other.is_external
            and self.is_secure == other.is_secure
            and self.volatility == other.volatility
            and self.is_null_call == other.is_null_call
            and self.comment == other.comment
            and self.is_external == other.is_external
            and self.api_integration == other.api_integration
            and self.context_headers == other.context_headers
            and self.max_batch_rows == other.max_batch_rows
            and self.compression == other.compression
        )


class InstanceProcedure(InstanceParameterObject):
    """
    Database object type Procedure
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.PROCEDURE, object_options)
        self.comment = metadata_query_result["COMMENT"]
        # the metadata that we currently query does not include information on EXECUTE AS
        # therefore, we can currently not use this class alone to determine if procedure instances are identical

    def __str__(self):
        return f"InstanceProcedure: {self.id}"

    def __repr__(self):
        return f"InstanceProcedure: {self.id}"

    def __eq__(self, other: "InstanceProcedure") -> bool:
        return super().__eq__(other) and self.comment == other.comment


class InstancePolicy(InstanceObject):
    """
    Database object type policy (masking or row access)
    """

    def __init__(self, metadata_query_result: Dict, object_type: DbObjectType, object_options: ObjectOption=None,):
        metadata_query_result["DATABASE_NAME"] = metadata_query_result["database_name"]
        metadata_query_result["SCHEMA_NAME"] = metadata_query_result["schema_name"]
        metadata_query_result["OBJECT_NAME"] = metadata_query_result["name"]
        super().__init__(metadata_query_result, object_type, object_options)
        self.kind = metadata_query_result["kind"]
        self.comment = metadata_query_result["comment"]
        self.signature = metadata_query_result["signature"]
        self.return_type = metadata_query_result["return_type"]
        self.body = metadata_query_result["body"]

    def __eq__(self, other: "InstancePolicy") -> bool:
        return (
            super().__eq__(other)
            and self.kind == other.kind
            and self.comment == other.comment
            and self.signature == other.signature
            and misc_utils.compare_datatype_ignore_varchar_maximum_length(self.return_type, other.return_type)
            and self.body == other.body
        )


class InstanceMaskingPolicy(InstancePolicy):
    """
    Database object type masking policy
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.MASKINGPOLICY, object_options)

    def __str__(self):
        return f"InstanceMaskingPolicy: {self.id}"

    def __repr__(self):
        return f"InstanceMaskingPolicy: {self.id}"


class InstanceRowAccessPolicy(InstancePolicy):
    """
    Database object type masking policy
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.ROWACCESSPOLICY, object_options)

    def __str__(self):
        return f"InstanceRowAccessPolicy: {self.id}"

    def __repr__(self):
        return f"InstanceRowAccessPolicy: {self.id}"


class InstanceSchema(InstanceObject):
    """
    Database object schema
    """

    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        self.database_name = metadata_query_result["DATABASE_NAME"]
        self.schema = metadata_query_result["SCHEMA_NAME"]
        self.name = self.schema
        self.object_type = DbObjectType.SCHEMA
        # self.is_managed_access = metadata_query_result["IS_MANAGED_ACCESS"]
        self.is_transient = metadata_query_result["IS_TRANSIENT"]
        self.retention_time = metadata_query_result["RETENTION_TIME"]
        self.database_retention_time = metadata_query_result["DATABASE_RETENTION_TIME"]
        self.comment = metadata_query_result["COMMENT"]
        self.object_options = object_options

    def __str__(self):
        return f"InstanceSchema: {self.id}"

    def __repr__(self):
        return f"InstanceSchema: {self.id}"

    def __eq__(self, other: "InstanceSchema") -> bool:
        return (
            self.object_type == other.object_type
            and self.schema == other.schema
            and self.name == other.name
            # and self.is_managed_access == other.is_managed_access
            and self.is_transient == other.is_transient
            and self.retention_time == other.retention_time
            and self.database_retention_time == other.database_retention_time
            and self.comment == other.comment
        )

    @property
    def full_name(self):
        return self.schema.upper()

    @property
    def id(self):
        return f"{self.object_type} {self.schema.upper()}"


class AppliedPolicyLegacy(object):
    """
    Hold information on which policy is applied to an object
    """

    def __init__(self, metadata, policy_type: PolicyType):
        self.policy_type = policy_type
        self.policy_db = metadata["POLICY_DB"]
        self.policy_schema = metadata["POLICY_SCHEMA"]
        self.policy_name = metadata["POLICY_NAME"]
        self.policy_kind = metadata["POLICY_KIND"]
        self.ref_database_name = metadata["REF_DATABASE_NAME"]
        self.ref_schema_name = metadata["REF_SCHEMA_NAME"]
        self.ref_entity_name = metadata["REF_ENTITY_NAME"]
        self.ref_entity_domain = metadata["REF_ENTITY_DOMAIN"]

    @property
    def full_name(self):
        "Full name including database and schema"
        return f"{self.policy_db}.{self.policy_schema}.{self.policy_name}"

    @property
    def full_name_quoted(self):
        "Full name including database and schema with quoted identifiers"
        return f'"{self.policy_db}"."{self.policy_schema}"."{self.policy_name}"'

    @staticmethod
    def factory(metadata):
        if metadata["POLICY_KIND"] == "ROW_ACCESS_POLICY":
            return AppliedRowAccessPolicyLegacy(metadata)
        elif metadata["POLICY_KIND"] == "MASKING_POLICY":
            return AppliedMaskingPolicyLegacy(metadata)
        else:
            raise ValueError(
                f"Metadata for POLICY_KIND [ '{metadata['POLICY_KIND']}' ] cannot be parsed"
            )

class InstanceDynamicTable(InstanceTableLike):
    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None, ignore_column_order: bool = False):
        super().__init__(metadata_query_result, DbObjectType.DYNAMICTABLE, object_options, ignore_column_order)
        self.query_text = metadata_query_result["QUERY_TEXT"]
        self.refresh_mode = metadata_query_result["REFRESH_MODE"]
        self.target_lag = metadata_query_result["TARGET_LAG"]
        self.warehouse = metadata_query_result["WAREHOUSE"]
        self.retention_time = metadata_query_result["RETENTION_TIME"]
        self.schema_retention_time = metadata_query_result["SCHEMA_RETENTION_TIME"]
        self.comment = metadata_query_result["COMMENT"]
        self.tags = metadata_query_result.get("tags", {})

    def __str__(self):
        return f"InstanceDynamicTable: {self.id}"

    def __repr__(self):
        return f"InstanceDynamicTable: {self.id}"

    def __eq__(self, other: "InstanceDynamicTable") -> bool:
        return (
            super().__eq__(other)
            and self.query_text == other.query_text
            and self.refresh_mode == other.refresh_mode
            and self.target_lag == other.target_lag
            and self.warehouse == other.warehouse
            and self.retention_time == other.retention_time
            and self.schema_retention_time == self.schema_retention_time
            and self.comment == other.comment
            and self.tags == other.tags
        )

class InstanceNetworkRule(InstanceObject):
    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        super().__init__(metadata_query_result, DbObjectType.NETWORKRULE, object_options)
        self.type = metadata_query_result["TYPE"]
        self.value_list = metadata_query_result["VALUE_LIST"]
        self.mode = metadata_query_result["MODE"]
        self.comment = metadata_query_result["COMMENT"]

    def __str__(self):
        return f"InstanceNetworkRule: {self.id}"

    def __repr__(self):
        return f"InstanceNetworkRule: {self.id}"

    def __eq__(self, other: "InstanceNetworkRule") -> bool:
        return (
            super().__eq__(other)
            and self.type == other.type
            and self.value_list == other.value_list
            and self.mode == other.mode
            and self.comment == other.comment
        )

class AppliedRowAccessPolicyLegacy(AppliedPolicyLegacy):
    """
    Hold information on which row access policy is applied to an object
    """

    def __init__(self, metadata):
        super().__init__(metadata, PolicyType.ROWACCESS)
        self.ref_arg_column_names = json.loads(metadata["REF_ARG_COLUMN_NAMES"])

    @property
    def column_list_string(self):
        "String containing all columns (to be used in ALTER VIEW ADD POLICY). Example: '(COL1,COL2)'"
        return f"({','.join(self.ref_arg_column_names)})"


class AppliedMaskingPolicyLegacy(AppliedPolicyLegacy):
    """
    Hold information on which masking policy is applied to an object
    """

    def __init__(self, metadata):
        super().__init__(metadata, PolicyType.MASKING)
        self.ref_column_name = metadata["REF_COLUMN_NAME"]

class InstanceTag(InstanceObject):
    def __init__(self, metadata_query_result: Dict, object_options: ObjectOption=None,):
        self.object_type = DbObjectType.TAG
        self.object_options = object_options
        self.schema = metadata_query_result["schema_name"]
        self.name = metadata_query_result["name"]
        if metadata_query_result["allowed_values"]:
            self.allowed_values = metadata_query_result["allowed_values"][1:-1].replace('"','').split(',')
        else:
            self.allowed_values=[]

        self.comment = metadata_query_result["comment"]

    def __str__(self):
        return f"InstanceTag: {self.id}"

    def __repr__(self):
        return f"InstanceTag: {self.id}"

    def __eq__(self, other: "InstanceTag") -> bool:
        return (
            self.object_type == other.object_type
            and self.schema == other.schema
            and self.name == other.name
            and self.allowed_values == other.allowed_values
            and self.comment == other.comment
        )


class PolicyReference(object):
    def __init__(self, metadata_query_result: Dict, policy_type: PolicyType):
        # IDEA: use only relevant attributes here and move others to child classes
        self.policy_type = policy_type       
        self.policy_db = metadata_query_result['POLICY_DB']
        self.policy_schema = metadata_query_result['POLICY_SCHEMA']
        self.policy_name = metadata_query_result['POLICY_NAME']
        self.policy_kind = metadata_query_result['POLICY_KIND']
        self.ref_database_name = metadata_query_result['REF_DATABASE_NAME']
        self.ref_schema_name = metadata_query_result['REF_SCHEMA_NAME']
        self.ref_entity_name = metadata_query_result['REF_ENTITY_NAME']
        self.ref_entity_domain = metadata_query_result['REF_ENTITY_DOMAIN']
        self.ref_column_name = metadata_query_result['REF_COLUMN_NAME']
        self.ref_arg_column_names = metadata_query_result['REF_ARG_COLUMN_NAMES']
        self.ref_arg_column_names_dict = json.loads(metadata_query_result["REF_ARG_COLUMN_NAMES"] or "[]")
        self.tag_database = metadata_query_result['TAG_DATABASE']
        self.tag_schema = metadata_query_result['TAG_SCHEMA']
        self.tag_name = metadata_query_result['TAG_NAME']
        self.policy_status = metadata_query_result['POLICY_STATUS']
    
    def __eq__(self, other) -> bool:
        return (
            self.policy_schema == other.policy_schema
            and self.policy_name == other.policy_name
            and self.policy_kind == other.policy_kind
            and self.ref_schema_name == other.ref_schema_name
            and self.ref_entity_name == other.ref_entity_name
            and self.ref_entity_domain == other.ref_entity_domain
            and self.ref_column_name == other.ref_column_name
            and self.ref_arg_column_names == other.ref_arg_column_names
            and self.ref_arg_column_names_dict == other.ref_arg_column_names_dict
            and self.tag_database == other.tag_database
            and self.tag_schema == other.tag_schema
            and self.tag_name == other.tag_name
            and self.policy_status == other.policy_status
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.policy_schema,
                self.policy_name,
                self.policy_kind,
                self.ref_schema_name,
                self.ref_entity_name,
                self.ref_entity_domain,
                self.ref_column_name,
                self.ref_arg_column_names,
                self.tag_database,
                self.tag_schema,
                self.tag_name,
                self.policy_status,
            )
        )

class PolicyReferenceFactory(object):
    @staticmethod
    def factory(metadata_query_result: Dict) -> PolicyReference:
        mapping = {
            "ROW_ACCESS_POLICY": RowAccessPolicyReference,
            "MASKING_POLICY": MaskingPolicyReference,
        }
        return mapping[metadata_query_result["POLICY_KIND"]](metadata_query_result)


class RowAccessPolicyReference(PolicyReference):
    def __init__(self, metadata):
        super().__init__(metadata, PolicyType.ROWACCESS)

    def __hash__(self) -> int:
        return super().__hash__()

    @property
    def ref_arg_columns_string(self):
        "String containing all columns (to be used in ALTER ... ADD POLICY). Example: '(COL1,COL2)'"
        return f"({','.join(self.ref_arg_column_names_dict)})"
    
    def __eq__(self, other) -> bool:
        return (
            super().__eq__(other)
        )

class MaskingPolicyReference(PolicyReference):
    def __init__(self, metadata):
        super().__init__(metadata, PolicyType.MASKING)

    def __hash__(self) -> int:
        return super().__hash__()

    @property
    def conditional_columns_string(self):
        "String containing all conditional columns (to be used in ALTER COLUMN ... ADD POLICY). Example: '(COL1,COL2)'"
        if len(self.ref_arg_column_names_dict) == 0:
            raise ValueError(f"Conditional columns string can only be generated for conditional masking policies. [ '{self.policy_schema}.{self.policy_name}' ] does not appear to be a conditional masking policy.")
        return f"({','.join([self.ref_column_name] + self.ref_arg_column_names_dict)})"



import logging
import os
import json
from typing import Dict, List, Union

import aceutils.file_util as file_util
import aceutils.general_util as gen_util
from aceaccount.core.model_account_object_sql_entities import AccountObjectType
from aceutils.logger import LoggingAdapter
from aceaccount.core.model_account_object_instances import AccountObjectInstance

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class AccountSolutionClient(object):
    """
    Object that collects all functionality based around the physical account object solution folder.
    All operations and information based upon these files is stored here.
    """

    def __init__(
        self,
        enabled_object_types: list,
        source_files: list,
    ) -> None:
        """
            Init a new SolutionClient
        Args:
            source_files: list - list of paths of the solution definitions
        """
        self._enabled_object_types = enabled_object_types
        self.source_files = source_files
        self.all_account_objects: List[AccountObjectInstance] = []
        self._load_solutions()

    def _dict_raise_on_duplicates(schema, ordered_pairs):
        """Reject duplicate keys."""
        d = {}
        for k, v in ordered_pairs:
            if k in d:
                raise ValueError("duplicate key: %r" % (k,))
            else:
                d[k] = v
        return d

    def _load_solutions(
        self,
    ) -> None:
        """
        Load the solutions form the source files
        """
        log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        # validate source file paths
        for source_file in self.source_files:
            if not os.path.isfile(source_file):
                raise EnvironmentError(
                    f"Solution path [ '{source_file}' ] is not valid"
                )

        mapping = {  # maps the name/key of the dict entry in the source file to the AccountObjectType
            "storage_integrations": AccountObjectType.STORAGEINTEGRATION,
            "warehouses": AccountObjectType.WAREHOUSE,
            "shares": AccountObjectType.SHARE,
            "databases": AccountObjectType.DATABASE,
            "external_volumes": AccountObjectType.EXTERNALVOLUME,
        }
        
        # load source files
        loaded_source_files = []
        
        for source_file in self.source_files:
            log.info(f"LOAD solution from [ '{source_file}' ]")

            source_dict = json.loads(
                file_util.load(source_file),
                object_pairs_hook=self._dict_raise_on_duplicates,
            )

            source_file_account_objects=[]

            for object_type_name in source_dict:
                if not object_type_name == "$schema":
                    if object_type_name not in mapping:
                        raise ValueError(
                            f"Error loading the solution: '{object_type_name}' is not in the list of the following supported object types: {[k for k in mapping]}"
                        )
                    
                    if mapping[object_type_name].value not in self._enabled_object_types:
                        log.debug(
                            f"SKIPPING disabled object type: {mapping[object_type_name].value}"
                        )
                        continue

                    for name, definition in source_dict[object_type_name].items():
                        account_object_instance = self._generate_account_object_instance(mapping[object_type_name], name, definition)
                        if not account_object_instance.id in [account_object.id for account_object in self.all_account_objects]:
                            source_file_account_objects.append(account_object_instance)
                            self.all_account_objects.append(account_object_instance)
                        else:
                            raise ValueError(
                                f"Error loading the solution {source_file} -> duplicate found: \n Account object of type '{object_type_name}' and name '{name}' already loaded. \n Please also check the previously loaded files to find the duplicate -> {loaded_source_files}"
                            )

            log.info(
                f"FOUND objects in [ '{source_file}' ]: {gen_util.generate_account_object_log_summaries(source_file_account_objects)}"
            )
            loaded_source_files.append(source_file)

        log.info(
            f"FOUND objects in all input files: {gen_util.generate_account_object_log_summaries(self.all_account_objects)}"
        )
        log.info("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
        

    def _generate_account_object_instance(self, object_type, name, definition):
        definition["name"] = name
        return AccountObjectInstance.factory(
            object_type=object_type,
            metadata_query_result=definition,
        )

    def get_object_by_object(self, obj: AccountObjectInstance) -> AccountObjectInstance:
        """
        Return the matching AccountObjectInstance for the given obj.
        Compares the id of the object.
        """
        return self.get_object_by_id(obj.id)

    def get_object_by_id(self, object_id: str) -> AccountObjectInstance:
        """
            Return the AccountObjectInstance for the given object_id.
        Args:
            id: str
        """
        result = [o for o in self.all_account_objects if (o.compare_id(object_id))]
        if len(result) == 0:
            log.debug(f"OBJECT with [ id='{object_id}' ] NOT in FOUND in solution")
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with [ id='{object_id}' ] NOT UNIQUE in solution. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]



import datetime
import logging
from typing import List, Set

import aceservices.snowflake_service as snowflake_service
import aceutils.dict_and_list_util as dict_and_list_util
import aceutils.file_util as file_util
import networkx as nx
from acedeploy.core.model_sql_entities import DbObjectType
from acedeploy.services.dependency_parser import DependencyParser
from acedeploy.services.solution_service import SolutionClient
from aceutils.logger import LoggingAdapter


logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def get_dependency_parser(
    solution_folder: str,
    pre_deployment_folders: List[str] = [],
    post_deployment_folders: List[str] = [],
    object_types: Set[DbObjectType] = (
        DbObjectType.TABLE,
        DbObjectType.VIEW,
        DbObjectType.MATERIALIZEDVIEW,
        DbObjectType.FUNCTION,
    ),
) -> DependencyParser:
    """
        Convienience function to gerenate a dependency parser client for a given folder
    Args:
        project_folder: str - path of solution to be loaded
        pre_deployment_folders: List[str] - predeployment folders must be given so they can be skipped
        post_deployment_folders: List[str] - postdeployment folders must be given so they can be skipped
        object_types: object_types: set[DbObjectType] - types of objects to be considered
    Returns:
        dependency_graph: DependencyParser - dependency parser
    """
    solution_client = SolutionClient(
        solution_folder, pre_deployment_folders, post_deployment_folders
    )
    solution_client.load_solution()
    dependency_client = DependencyParser(solution_client)
    dependency_client.build_full_dependency_graph(object_types)
    return dependency_client


def get_dependency_graph(
    solution_folder: str,
    pre_deployment_folders: List[str] = [],
    post_deployment_folders: List[str] = [],
    object_types: Set[DbObjectType] = (
        DbObjectType.TABLE,
        DbObjectType.VIEW,
        DbObjectType.MATERIALIZEDVIEW,
        DbObjectType.FUNCTION,
    ),
) -> nx.DiGraph:
    """
        Convienience function to gerenate a dependency graph for a given folder
    Args:
        project_folder: str - path of solution to be loaded
        pre_deployment_folders: List[str] - predeployment folders must be given so they can be skipped
        post_deployment_folders: List[str] - postdeployment folders must be given so they can be skipped
        object_types: object_types: set[DbObjectType] - types of objects to be considered
    Returns:
        dependency_graph: nx.DiGraph - dependency graph
    """
    dependency_client = get_dependency_parser(
        solution_folder, pre_deployment_folders, post_deployment_folders, object_types
    )
    return dependency_client._dependency_graph


def convert_dependency_graph_to_nested_dict(
    dependency_graph: nx.DiGraph, schema_filter: List[str] = []
) -> dict:
    """
        Convert a given dependency graph to a nested dictionary
    Args:
        dependency_graph: nx.DiGraph - dependency graph of solution objects
        schema_filter: List[str] - only get information on objects in these schemas (empty=all schemas)
    Returns:
        dict - dictionary containing information on dependency graph
    """
    log.info(f"CONVERT graph to nested DICT")
    if schema_filter:
        schema_filter_upper = [s.upper() for s in schema_filter]
        nodes = [n for n in dependency_graph.nodes() if n.schema in schema_filter_upper]
    else:
        nodes = list(dependency_graph.nodes())
    result = []

    for node in nodes:
        result.append(
            {
                "type": DbObjectType.get_sql_object_type(node.object_type),
                "schema": node.schema,
                "name": node.name,
                "dependsOn": [
                    {
                        "type": DbObjectType.get_sql_object_type(n.object_type),
                        "schema": n.schema,
                        "name": n.name,
                    }
                    for n in nx.descendants(dependency_graph.reverse(), node)
                ],
                "requiredFor": [
                    {
                        "type": DbObjectType.get_sql_object_type(n.object_type),
                        "schema": n.schema,
                        "name": n.name,
                    }
                    for n in nx.descendants(dependency_graph, node)
                ],
            }
        )

    return result


def convert_dependency_graph_to_nested_json(
    file_path: str, dependency_graph: nx.DiGraph, schema_filter: List[str] = []
) -> None:
    """
        Convert a given dependency graph to a nested json
    Args:
        file_path: str - path where the output will be stored (overwritten without asking)
        dependency_graph: nx.DiGraph - dependency graph of solution objects
        schema_filter: List[str] - only get information on objects in these schemas (empty=all schemas)
    Returns:
        str - json containing information on dependency graph
    """
    dictionary = convert_dependency_graph_to_nested_dict(
        dependency_graph, schema_filter
    )
    log.info(f"SAVE graph as [ '{file_path}' ]")
    file_util.save_json(file_path, dictionary)


def convert_dependency_graph_to_list_of_dict(
    dependency_graph: nx.DiGraph, schema_filter: List[str] = []
) -> List[dict]:
    """
        Convert a given dependency graph to a list of dictionaries
    Args:
        dependency_graph: nx.DiGraph - dependency graph of solution objects
        schema_filter: List[str] - only get information on objects in these schemas (empty=all schemas)
    Returns:
        List[dict] - list containing information on dependency graph
    """
    log.info(f"CONVERT graph to LIST of DICTS")
    if schema_filter:
        schema_filter_upper = [s.upper() for s in schema_filter]
        nodes = [n for n in dependency_graph.nodes() if n.schema in schema_filter_upper]
    else:
        nodes = list(dependency_graph.nodes())
    result = []

    for node in nodes:
        result.extend(
            [
                {
                    "type": DbObjectType.get_sql_object_type(node.object_type),
                    "schema": node.schema,
                    "name": node.name,
                    "dependsOnType": DbObjectType.get_sql_object_type(n.object_type),
                    "dependsOnSchema": n.schema,
                    "dependsOnName": n.name,
                }
                for n in nx.descendants(dependency_graph.reverse(), node)
            ]
        )

    return result


def convert_dependency_graph_to_flat_json(
    file_path: str, dependency_graph: nx.DiGraph, schema_filter: List[str] = []
) -> None:
    """
        Convert a given dependency graph to a flat json
    Args:
        file_path: str - path where the output will be stored (overwritten without asking)
        dependency_graph: nx.DiGraph - dependency graph of solution objects
        schema_filter: List[str] - only get information on objects in these schemas (empty=all schemas)
    Returns:
        str - json containing information on dependency graph
    """
    list_of_dicts = convert_dependency_graph_to_list_of_dict(
        dependency_graph, schema_filter
    )
    log.info(f"SAVE graph as [ '{file_path}' ]")
    file_util.save_json(file_path, list_of_dicts)


def convert_dependency_graph_to_csv(
    file_path: str,
    dependency_graph: nx.DiGraph,
    schema_filter: List[str] = [],
    csv_delimiter: str = ";",
) -> None:
    """
        Convert a given dependency graph to a csv
    Args:
        file_path: str - path where the output will be stored (overwritten without asking)
        dependency_graph: nx.DiGraph - dependency graph of solution objects
        schema_filter: List[str] - only get information on objects in these schemas (empty=all schemas)
    Returns:
        str - json containing information on dependency graph
    """
    list_of_dicts = convert_dependency_graph_to_list_of_dict(
        dependency_graph, schema_filter
    )
    log.info(f"SAVE graph as [ '{file_path}' ]")
    file_util.save_list_of_dicts_as_csv(
        file_path, list_of_dicts, delimiter=csv_delimiter
    )


def write_dependencies_to_table(
    snowclient_config: snowflake_service.SnowClientConfig,
    snowflake_database: str,
    snowflake_schema: str,
    snowflake_table: str,
    dependency_graph: nx.DiGraph,
    schema_filter: List[str] = [],
) -> None:
    """
    Write information in given dependency graph to a Snowflake table.
    The table must exist. Table will be truncated.

    DDL of the target table:
        CREATE OR REPLACE TABLE <database>.PUBLIC.OBJECT_DEPENDENCIES (
            TYPE VARCHAR COMMENT 'Type of the object',
            SCHEMA VARCHAR COMMENT 'Schema of the object',
            NAME VARCHAR COMMENT 'Name of the object',
            DEPENDS_ON_TYPE VARCHAR COMMENT 'Type of the object that this object depends on',
            DEPENDS_ON_SCHEMA VARCHAR COMMENT 'Type of the object that this object depends on',
            DEPENDS_ON_NAME VARCHAR COMMENT 'Type of the object that this object depends on',
            UPDATE_TIMESTAMP TIMESTAMP_TZ COMMENT 'Time at which this entry has been generated'
        );
    """
    timestamp = datetime.datetime.now(datetime.timezone.utc)
    dependency_dicts_list = convert_dependency_graph_to_list_of_dict(
        dependency_graph, schema_filter
    )
    statements = _generate_statements_snowflake_table(
        dependency_dicts_list,
        timestamp,
        snowflake_database,
        snowflake_schema,
        snowflake_table,
    )
    _execute_on_snowflake(statements, snowclient_config)


def _generate_statements_snowflake_table(
    dependency_dicts_list,
    timestamp,
    snowflake_database,
    snowflake_schema,
    snowflake_table,
) -> str:
    """
    Generate statements to insert the dependency info into a table.
    """
    log.info(f"GENERATE sql statements")

    statements = ["BEGIN;"]
    statements.append(
        f"TRUNCATE {snowflake_database}.{snowflake_schema}.{snowflake_table};"
    )

    values = []
    for d in dependency_dicts_list:
        values.append(
            f"""(
                '{d['type']}',
                '{d['schema']}',
                '{d['name']}',
                '{d['dependsOnType']}',
                '{d['dependsOnSchema']}',
                '{d['dependsOnName']}',
                TO_TIMESTAMP_TZ('{timestamp}')
            )"""
        )
    for value_sublist in dict_and_list_util.chunks(
        values, 10000
    ):  # insert statements cannot contain more than 16384 rows
        insert_statement = f"""
            INSERT INTO {snowflake_database}.{snowflake_schema}.{snowflake_table} (
                TYPE,
                SCHEMA,
                NAME,
                DEPENDS_ON_TYPE,
                DEPENDS_ON_SCHEMA,
                DEPENDS_ON_NAME,
                UPDATE_TIMESTAMP
            ) VALUES
            {','.join(value_sublist)}
            ;"""
        statements.append(insert_statement)

    log.debug(".   \n".join(statements))

    statements.append("COMMIT;")

    return statements


def _execute_on_snowflake(
    statements: str, snowclient_config: snowflake_service.SnowClientConfig
):
    """
    Connect to snowflake and execute a set of statements.
    """
    log.info(f"CONNECT to Snowflake")
    snowflake_client = snowflake_service.SnowClient(snowclient_config)
    snowflake_client.execute_statement(f"USE WAREHOUSE {snowclient_config.warehouse};")
    log.info(f"EXECUTE sql statements")
    snowflake_client.execute_statement(statements)



from typing import  Dict, Tuple
import copy

def compare_dicts(dict1, dict2):
    """
    Compare two dictionaries. Dictionaries may only contain hashable types.
    Return set of two dictionaries.
    First dictionary contains all items only in first input dictionary.
    Second dictionary contains all items only in second input dictionary.
    """
    set1 = set(dict1.items())
    set2 = set(dict2.items())
    only_in_dict1 = dict(set1 - set2)
    only_in_dict2 = dict(set2 - set1)
    return only_in_dict1, only_in_dict2


def compare_nested_dicts(dict1, dict2) -> Tuple[Dict,Dict]:
    """
    Compare two nested dictionaries - only one sublayer supported. Dictionaries may only contain hashable types or "non-nested"-dictionaries.
    Return two nested dictionaries.
    First nested dictionary contains all items only in first input nested dictionary.
    Second nested dictionary contains all items only in second input nested dictionary.
    """
    dict1_sub_diffs={}
    dict2_sub_diffs={}

    dict1_copy=copy.deepcopy(dict1)
    dict2_copy=copy.deepcopy(dict2)

    for key, value in dict1_copy.items():
        if type(value) is dict and key not in dict2_copy:
            dict1_sub_diffs[key]=value
            dict1.pop(key)
        elif type(value) is dict and type(dict2_copy[key]) is not dict:
            raise ValueError(f"Nested dictionaries don't have the same structure.")
        elif type(value) is dict and type(dict2_copy[key]) is dict:
            sub_dict_1=dict1.pop(key)
            sub_dict_2=dict2.pop(key)

            only_in_sub_dict1, only_in_sub_dict2 = compare_dicts(sub_dict_1, sub_dict_2)

            if only_in_sub_dict1:
                dict1_sub_diffs[key]=only_in_sub_dict1
            if only_in_sub_dict2:
                dict2_sub_diffs[key]=only_in_sub_dict2

    for key, value in dict2_copy.items():
        if type(value) is dict and key not in dict1_copy:
            dict2_sub_diffs[key]=value
            dict2.pop(key)
        elif type(value) is dict and type(dict1_copy[key]) is not dict:
            raise ValueError(f"Nested dictionaries don't have the same structure.")

    only_in_dict1, only_in_dict2 = compare_dicts(dict1, dict2)
    only_in_dict1 |= dict1_sub_diffs 
    only_in_dict2 |= dict2_sub_diffs 
    return only_in_dict1, only_in_dict2

def strip_nested_dict(d):
    """
    Remove empty entries from a (nested) dictioary.
    Entries count as empty if they are one of ('', None, [], {}).
    """
    result = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = strip_nested_dict(v)
        if v not in ("", None, [], {}):
            result[k] = v
    return result


def list_casefold(string_list):
    """
    Given a list of string, apply casefold() to each of them.
    If the input is a Null value, return empty list.
    """
    if string_list:
        return [s.casefold() for s in string_list]
    else:
        return []


def list_upper(string_list):
    """
    Given a list of string, apply upper() to each of them.
    If the input is a Null value, return empty list.
    """
    if string_list:
        return [s.upper() for s in string_list]
    else:
        return []


def list_lower(string_list):
    """
    Given a list of string, apply lower() to each of them.
    If the input is a Null value, return empty list.
    """
    if string_list:
        return [s.lower() for s in string_list]
    else:
        return []


def flatten(nested_list):
    """
    Flatten arbitrary nested lists
    """
    result = []
    if isinstance(nested_list, (list, tuple)):
        for x in nested_list:
            result.extend(flatten(x))
    else:
        result.append(nested_list)
    return result


def chunks(lst, n):
    """
    Yield successive n-sized chunks from lst.
    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]

def list_to_string_representation(lst) -> str:
    """
    Transforms a list to a string representation which can, for example, be used in Snowflake.
    E.g. ["a","b"] -> "'a','b'"
    """
    result = ", ".join(
        [f"'{s}'" for s in lst]
    )
    return result

def list_ends_with_list(list_a, list_b):
    """
    Check if list A ends with entries given in list B.
    If list B is empty, return True.
    If list B is longer than list A, return False.
    """
    if len(list_b) == 0:
        return True
    if len(list_b) > len(list_a):
        return False
    return list_a[-len(list_b):] == list_b



# These functions provide utilites for working with directed graphs (nx.DiGraph).

import networkx as nx


def remove_self_loops(graph):
    """
    Remove all self looping edges from a graph

    See https://stackoverflow.com/questions/49427638/removing-self-loops-from-undirected-networkx-graph
    """
    new_graph = graph.copy()
    new_graph.remove_edges_from(nx.selfloop_edges(new_graph))
    return new_graph


def split_graph_into_subgraphs(graph):
    """
    Return subgraphs of a graph, i.e. no subgraph shares an edge or node with any other subgraph.
    """
    return [graph.subgraph(c).copy() for c in nx.weakly_connected_components(graph)]


def get_all_successor_nodes(graph, node):
    """
    Get a set of all nodes which succeed the given node, including the given node itself.
    """
    nodes = nx.descendants(graph, source=node)
    nodes.add(node)
    return nodes


def get_successor_graph(graph, node):
    """
    Get a subgraph of the given graph, containing only the given node and its successors.
    """
    nodes = get_all_successor_nodes(graph, node)
    return graph.subgraph(nodes)


def get_all_predecessor_nodes(graph, node):
    """
    Get a set of all nodes which preceed the given node, including the given node itself.
    """
    nodes = nx.ancestors(graph, source=node)
    nodes.add(node)
    return nodes


def get_predecessor_graph(graph, node):
    """
    Get a subgraph of the given graph, containing only the given node and its predecessors.
    """
    nodes = get_all_predecessor_nodes(graph, node)
    return graph.subgraph(nodes)


def get_ordered_objects(graph):
    """
    Get a list of all objects in a given graph, ordered topologically.
    """
    return list(nx.topological_sort(graph))


def filter_graph(graph, nodes, mode):
    """
    Filter a graph to contain only the given nodes and additional nodes, depending on mode.
    mode = 'given_nodes_only', 'include_predecessors', 'include_successors', 'include_both'
    See docstring of called functions for details.
    """
    if mode.lower() == "given_nodes_only":
        return filter_graph_given_only(graph, nodes)
    elif mode.lower() == "include_predecessors":
        return filter_graph_include_predecessors(graph, nodes)
    elif mode.lower() == "include_successors":
        return filter_graph_include_successors(graph, nodes)
    elif mode.lower() == "include_both":
        return filter_graph_include_both(graph, nodes)
    else:
        raise ValueError(
            f"Given mode [ '{mode}' ] not supported. Supported modes are 'given_nodes_only', 'include_predecessors', 'include_successors', 'include_both'."
        )


def filter_graph_given_only(graph, nodes):
    """
    Filter a graph to contain only the given nodes.
    """
    return graph.subgraph(nodes).copy()


def filter_graph_include_predecessors(graph, nodes):
    """
    Filter a graph to contain only the given nodes and any nodes which preceed these nodes.
    """
    predecessor_nodes = set()
    for n in nodes:
        if n not in predecessor_nodes:
            predecessor_nodes.update(get_all_predecessor_nodes(graph, n))
    return graph.subgraph(predecessor_nodes).copy()


def filter_graph_include_successors(graph, nodes):
    """
    Filter a graph to contain only the given nodes and any nodes which succeed these nodes.
    """
    successor_nodes = set()
    for n in nodes:
        if n not in successor_nodes:
            successor_nodes.update(get_all_successor_nodes(graph, n))
    return graph.subgraph(successor_nodes).copy()


def filter_graph_include_both(graph, nodes):
    """
    Filter a graph to contain only the given nodes and any nodes which succeed or preceed these nodes.
    """
    return nx.compose(
        filter_graph_include_predecessors(graph, nodes),
        filter_graph_include_successors(graph, nodes),
    )



from typing import List, Tuple

from acedeploy.core.model_database_object import DatabaseObject
from aceaccount.core.model_account_object import (
    AccountObject,
    AccountObjectType,
)
from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
import aceutils.file_util as file_util


def generate_action_log_summaries(
    action_list: List[DbObjectAction],
) -> Tuple[str, str, str]:
    """
        Generate a short summary of actions to perform.
    Args:
        action_list: List[DbObjectAction] - list of db object actions
    Returns:
        string_add, string_alter, string_drop: (str, str, str) - each e.g. "create: 3 schema(s), 1 table(s), 2 view(s), ..."
    """
    dropoverloaded = False
    counts_add = {object_type: 0 for object_type in [o.value for o in DbObjectType]}
    counts_alter = {object_type: 0 for object_type in [o.value for o in DbObjectType]}
    counts_drop = {object_type: 0 for object_type in [o.value for o in DbObjectType]}
    counts_dropoverloaded = {object_type: 0 for object_type in [DbObjectType.PROCEDURE.value, DbObjectType.FUNCTION.value]}
    for action in action_list:
        if action.action == DbActionType.ADD:
            counts_add[action.object_type.value] += 1
        elif action.action == DbActionType.ALTER:
            counts_alter[action.object_type.value] += 1
        elif action.action == DbActionType.DROP:
            counts_drop[action.object_type.value] += 1
        elif action.action == DbActionType.DROPOVERLOADED:
            counts_dropoverloaded[action.object_type.value] += 1
            dropoverloaded = True
    
    if dropoverloaded:
        string_dropoverloaded = "drop_overloaded: " + ", ".join(
            [f"{counts_dropoverloaded[key]} {key.lower()}(s)" for key in counts_dropoverloaded]
        )
    else:
        string_dropoverloaded = ""

    string_add = "create: " + ", ".join(
        [f"{counts_add[key]} {key.lower()}(s)" for key in counts_add]
    )
    string_alter = "alter: " + ", ".join(
        [f"{counts_alter[key]} {key.lower()}(s)" for key in counts_alter]
    )
    string_drop = "drop: " + ", ".join(
        [f"{counts_drop[key]} {key.lower()}(s)" for key in counts_drop]
    )
    return string_dropoverloaded, string_add, string_alter, string_drop


def generate_object_log_summaries(object_list: List[DatabaseObject]) -> str:
    """
        Generate a short summary of given objects.
    Args:
        object_list: List[DatabaseObject] - list of db objects
    Returns:
        string - e.g. "create: 3 schema(s), 1 table(s), 2 view(s), ..."
    """
    counts = {object_type: 0 for object_type in [o.value for o in DbObjectType]}
    for obj in object_list:
        counts[obj.object_type.value] += 1
    return ", ".join([f"{counts[key]} {key.lower()}(s)" for key in counts])


def generate_account_object_log_summaries(object_list: List[AccountObject]) -> str:
    """
        Generate a short summary of given account objects.
    Args:
        object_list: List[AccountObject] - list of account objects
    Returns:
        string - e.g. "create: 3 schema(s), 1 table(s), 2 view(s), ..."
    """
    counts = {object_type: 0 for object_type in [o.value for o in AccountObjectType]}
    for obj in object_list:
        counts[obj.object_type.value] += 1
    return ", ".join([f"{counts[key]} {key.lower()}(s)" for key in counts])


def save_action_list(
    path: str,
    action_list: List[DbObjectAction],
    predeployment_steps: List[str],
    postdeployment_steps: List[str],
):
    """Save action list in file"""
    with open(path, "w") as f:
        action_summary = "\n".join(generate_action_log_summaries(action_list))
        f.write(f"{action_summary}\n")
        f.write("\nList of actions: \n")
        if len(action_list) == 0:
            f.write("(none)")
        for action in action_list:
            text = (
                f"{action.action.name} {action.object_type.name} {action.full_name}\n"
            )
            f.write(text)

        pre_steps = [s for s in predeployment_steps if s.execute_step]
        f.write("\nList of predeployment steps: \n")
        if len(pre_steps) == 0:
            f.write("(none)")
        for step in pre_steps:
            f.write(f"\n{step.path}\n{step.content}\n")

        post_steps = [s for s in postdeployment_steps if s.execute_step]
        f.write("\nList of postdeployment steps: \n")
        if len(post_steps) == 0:
            f.write("(none)")
        for step in post_steps:
            f.write(f"\n{step.path}\n{step.content}\n")


def save_action_json(path, action_list, database, role):
    """
    Save action json to file
    """
    json_object = {}
    json_object["database"] = database
    json_object["role"] = role
    json_object["objects"] = {}
    json_object["objects"]["views"] = [
        a.full_name for a in action_list if a.object_type == DbObjectType.VIEW
    ]
    json_object["objects"]["tables"] = [
        a.full_name for a in action_list if a.object_type == DbObjectType.TABLE
    ]

    file_util.save_json(path, json_object)



import logging
from typing import List

import matplotlib
import networkx as nx
from acedeploy.services.dependency_parser import DependencyParser
from aceutils.logger import LoggingAdapter

matplotlib.use("Agg")

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def plot_dependency_graph(
    file_path: str, dependency_client: DependencyParser, object_filter: List[str] = []
) -> None:
    """
    Plot a given dependency graph.

    Plot either the full graph or parts of the graph.
    When plotting parts of the graph, show all objects that a given object depends on and objects that depend on the given object.

    file_path: str - Path where the plot will be stored
    dependency_client: DependencyParser - parser object containing the dependency graph
    object_filter: List[str] - show dependencies only for given full object names ('<schema>.<name>'). If an empty list is supplied, plot the complete graph.
    """
    if object_filter:
        dependency_client.filter_graph_by_object_names(object_filter, "plot")
    dependency_graph = dependency_client._dependency_graph

    pos = nx.shell_layout(dependency_graph.reverse())
    # nx.draw_networkx_nodes(dependency_graph, pos, node_size=300)
    nx.draw_networkx_edges(dependency_graph.reverse(), pos, edge_color=".6")
    labels = {
        s: s.full_name.replace(".", ".\n") for s in dependency_graph.reverse().nodes()
    }
    nx.draw_networkx_labels(dependency_graph.reverse(), pos, labels=labels, font_size=6)

    log.info(f"SAVE graph as [ '{file_path}' ]")
    matplotlib.pyplot.savefig(file_path)



import logging
import re
from typing import List

import networkx as nx

import aceutils.graphtools as graphtools
import aceutils.string_util as string_util
from acedeploy.core.model_solution_entities import SolutionFunction, SolutionObject
from acedeploy.core.model_sql_entities import DbFunctionType, DbObjectType
from acedeploy.services.solution_service import SolutionClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class DependencyParser(object):
    """
    Parser which enables searching for dependencies in sql DDL statements.
    """

    def __init__(self, solution_client: SolutionClient):
        """
            Inits a new dependency parser client
        Args:
            solution_client: SolutionClient
        """
        self._solution_client = solution_client
        self._dependency_graph = nx.DiGraph()
        self._dependency_subgraphs = []

    def build_full_dependency_graph(
        self,
        object_types=(
            DbObjectType.VIEW,
            DbObjectType.MATERIALIZEDVIEW,
            DbObjectType.TABLE,
            DbObjectType.EXTERNALTABLE,
            DbObjectType.FUNCTION,
            DbObjectType.PROCEDURE,
            DbObjectType.FILEFORMAT,
            DbObjectType.SEQUENCE,
            DbObjectType.STAGE,
            DbObjectType.PIPE,
            DbObjectType.STREAM,
            DbObjectType.MASKINGPOLICY,
            DbObjectType.ROWACCESSPOLICY,
            DbObjectType.TASK,
            DbObjectType.DYNAMICTABLE,
            DbObjectType.HYBRIDTABLE,
            DbObjectType.NETWORKRULE
        ),
    ):
        """
        Build a dependency graph for all objects in the solution client
        """
        relevant_objects = [
            o
            for o in self._solution_client.all_objects
            if o.object_type in object_types
        ]
        self.function_name_list = [
            f"{o.schema}.{o.name}"
            for o in relevant_objects
            if o.object_type == DbObjectType.FUNCTION
        ]
        log.info(f"BUILD dependency GRAPH for [ '{len(relevant_objects)}' ] objects")
        for obj in relevant_objects:
            self._dependency_graph.add_node(obj)
            dependencies = self._get_dependencies(obj)
            for dependency in dependencies:
                log.debug(f"DEPENDENCY for [ '{obj.id}' ] FOUND [ '{dependency.id}' ]")
                self._dependency_graph.add_edge(dependency, obj)
        self._dependency_graph = graphtools.remove_self_loops(self._dependency_graph)

    def _get_dependencies(self, obj: SolutionObject):
        """
        Given an SolutionObject, return all objects that this object depends on.
        """
        dependencies_objects_names = self._parse_object_dependencies(
            obj.content, obj.object_type
        )
        dependencies_objects = []
        for dependency_object_name in dependencies_objects_names:
            dependency = self._solution_client.get_object_by_full_name(
                full_name=dependency_object_name,
                include_schemas=False,
            )
            if dependency:
                dependencies_objects.append(dependency)
            else:
                dependency = (
                    self._solution_client.get_parameterobject_by_name_without_params(
                        dependency_object_name
                    )
                )
                if dependency:
                    dependencies_objects.append(dependency)
                else:
                    log.info(
                        f"DEPENDENCY [ '{dependency_object_name}' ] for [ '{obj.id}' ] NOT in solution"
                    )

        return dependencies_objects

    def filter_graph_by_object_ids(self, object_ids, mode):
        """
        Given a list of object IDs, remove all objects from the graph which are not required for the creation of those objects.
        mode = 'meta_deployment' selects objects required for meta deployment.
        mode = 'target_deployment' selects objects required for target deployment.
        mode = 'plot' selects objects for plotting.
        """
        objects = [self._solution_client.get_object_by_id(oid) for oid in object_ids]
        log.info(
            f"FILTER dependency GRAPH with MODE [ '{mode}' ] for [ '{len(object_ids)}' ] objects"
        )
        self.filter_graph(objects, mode)

    def filter_graph_by_object_names(self, full_object_names, mode):
        """
        Given a list of full object names, remove all objects from the graph which are not required for the creation of those objects.
        mode = 'meta_deployment' selects objects required for meta deployment.
        mode = 'target_deployment' selects objects required for target deployment.
        mode = 'plot' selects objects for plotting.
        """
        objects = [
            self._solution_client.get_object_by_full_name(
                full_name=full_name,
                include_schemas=False,
            )
            for full_name in full_object_names
        ]
        log.info(
            f"FILTER dependency GRAPH with MODE [ '{mode}' ] for [ '{len(full_object_names)}' ] objects"
        )
        self.filter_graph(objects, mode)

    def filter_graph(self, objects, mode):
        """
        Given a list of objects, remove all objects from the graph which are not required for the creation of those objects.
        mode = 'meta_deployment' selects objects required for meta deployment.
        mode = 'target_deployment' selects objects required for target deployment.
        mode = 'plot' selects objects for plotting.
        """
        if set(objects) == set(self._dependency_graph.nodes()):
            # no need to filter if all nodes will be used
            log.debug("SKIP filter dependency graph: no objects need to be removed")
        elif mode.lower() == "meta_deployment":
            log.debug(f"GET dependency graph succesors for [ {len(objects)} ] objects")
            dependency_graph_successors = graphtools.filter_graph(
                self._dependency_graph, objects, "include_successors"
            )
            successor_nodes = [n for n in dependency_graph_successors.nodes()]
            log.debug(
                f"GET dependency graph predecessors for [ {len(successor_nodes)} ] objects"
            )
            self._dependency_graph = graphtools.filter_graph(
                self._dependency_graph, successor_nodes, "include_predecessors"
            )
        elif mode.lower() == "target_deployment":
            # in a target deployment, we assume that 'objects' contains all views that were deployed in the meta deployment.
            # therefore, it is not necessary to add successors in a target deployment, since all views that might need to be
            # updated already appear in 'objects'.
            log.debug(
                f"GET dependency graph for given objects ([ {len(objects)} ] objects)"
            )
            self._dependency_graph = graphtools.filter_graph(
                self._dependency_graph, objects, "given_nodes_only"
            )
        elif mode.lower() == "plot":
            # when plotting a graph, we want to look at all objects that directly depend on the given object
            # and that are a prerequisite of a given object
            log.debug(f"GET dependency graph succesors for [ {len(objects)} ] objects")
            dependency_graph_successors = graphtools.filter_graph(
                self._dependency_graph, objects, "include_successors"
            )
            successor_nodes = [n for n in dependency_graph_successors.nodes()]
            log.debug(
                f"GET dependency graph predecessors for [ {len(objects)} ] objects"
            )
            dependency_graph_predecessors = graphtools.filter_graph(
                self._dependency_graph, objects, "include_predecessors"
            )
            predecessors_nodes = [n for n in dependency_graph_predecessors.nodes()]
            relevant_nodes = set().union(successor_nodes, predecessors_nodes)
            self._dependency_graph = graphtools.filter_graph(
                self._dependency_graph, relevant_nodes, "given_nodes_only"
            )
        else:
            raise ValueError(
                f"Mode [ '{mode}' ] is unknown. Valid modes: [ 'meta_deployment', 'target_deployment', 'plot' ]"
            )

    def build_subgraphs(self):
        """
        Build dependency subgraphs (graphs with no interdependencies).
        """
        log.info("BUILD independent SUBGRAPHS")
        self._dependency_subgraphs = graphtools.split_graph_into_subgraphs(
            self._dependency_graph
        )

    def get_ordered_objects(self):
        """
        For each subgraph, return a list of ordered objects.
        """
        log.info(f"ORDER [ '{len(self._dependency_subgraphs)}' ] SUBGRAPHS")
        return [graphtools.get_ordered_objects(g) for g in self._dependency_subgraphs]

    def _parse_object_dependencies(
        self, statement: str, object_type: DbObjectType
    ) -> List[str]:
        """
        For a given statement and object type, return all dependent object names
        """
        mapping = {
            DbObjectType.SCHEMA: DependencyParser._dummy_parser_empty_result,
            DbObjectType.TABLE: DependencyParser._parse_object_dependencies_tables,
            DbObjectType.EXTERNALTABLE: DependencyParser._parse_object_dependencies_externaltables,
            DbObjectType.VIEW: self._parse_object_dependencies_views,
            DbObjectType.MATERIALIZEDVIEW: self._parse_object_dependencies_views,
            DbObjectType.FUNCTION: self._parse_object_dependencies_functions,
            DbObjectType.PROCEDURE: DependencyParser._dummy_parser_empty_result,
            DbObjectType.STAGE: DependencyParser._parse_object_dependencies_stages,
            DbObjectType.FILEFORMAT: DependencyParser._dummy_parser_empty_result,
            DbObjectType.STREAM: DependencyParser._parse_object_dependencies_streams,
            DbObjectType.TASK: DependencyParser._parse_object_dependencies_tasks,
            DbObjectType.PIPE: DependencyParser._parse_object_dependencies_pipes,
            DbObjectType.SEQUENCE: DependencyParser._dummy_parser_empty_result,
            DbObjectType.MASKINGPOLICY: self._parse_object_dependencies_policies,
            DbObjectType.ROWACCESSPOLICY: self._parse_object_dependencies_policies,
            DbObjectType.DYNAMICTABLE: self._parse_object_dependencies_dynamictables,
            DbObjectType.HYBRIDTABLE: self._parse_object_dependencies_hybridtables,
            DbObjectType.NETWORKRULE: self._dummy_parser_empty_result
        }
        return mapping[object_type](statement)

    @staticmethod
    def _dummy_parser_empty_result(*_, **__) -> List[str]:
        """
        For all objects without dependencies, return an empty list
        """
        return []

    def _parse_object_dependencies_functions(self, statement: str) -> List[str]:
        """
        Parses function dependencies from a given statement
        Will only parse for SQL UDFs, all other types do not need to be validated.
        """
        return DependencyParser._parse_object_dependencies_functions_static(
            statement, self.function_name_list
        )

    @staticmethod
    def _parse_object_dependencies_functions_static(
        statement: str, function_name_list: List[str]
    ) -> List[str]:
        """
        Parses function dependencies from a given statement
        Will only parse for SQL UDFs, all other types do not need to be validated.
        """
        if SolutionFunction.get_function_type(statement) in (DbFunctionType.SQL,):
            definition_regex = re.compile(
                r"(?:\$\$|')(?P<definition>.*)(?:\$\$|')", re.IGNORECASE | re.DOTALL
            )
            definition = definition_regex.search(statement).group("definition")
            return DependencyParser._parse_object_dependencies_views_functions_policies_static(
                definition, function_name_list
            )
        else:
            return []

    def _parse_object_dependencies_views(self, statement: str) -> List[str]:
        """
        Parses view dependencies from a given statement
        Will check for FROM statements, explicit and implicit joins and functions
        """
        return DependencyParser._parse_object_dependencies_views_static(
                statement, self.function_name_list
            )

    def _parse_object_dependencies_policies(self, statement: str) -> List[str]:
        """
        Parses policy dependencies from a given statement
        Will check for FROM statements, explicit and implicit joins and functions
        """
        return (
            DependencyParser._parse_object_dependencies_views_functions_policies_static(
                statement, self.function_name_list
            )
        )

    @staticmethod
    def _parse_object_dependencies_views_static(
        statement: str, function_name_list: List[str] = None
    ) -> List[str]:
        """
            Parses view dependencies from a given statement
            Will check for FROM statements, explicit and implicit joins, and policies
        Args:
            statement: str - the whole sql statement (multiple lines)
            function_name_list: List[str] - (optional) list of function names which might be a dependency for this statement
        """
        if function_name_list is None:
            function_name_list = []

        dependencies = []
        row_access_policy_pattern = re.compile(
            r"(?:[\s\)])ROW\s+ACCESS\s+POLICY\s+(?P<referenced_policy>[\w\.\"]+)\s+ON\s*",
            re.IGNORECASE | re.DOTALL,
        )
        row_access_policy_matches = row_access_policy_pattern.findall(statement)
        dependencies.extend(row_access_policy_matches)

        masking_policy_pattern = re.compile(
            r"(?:[\s\)])MASKING\s+POLICY\s+(?P<referenced_policy>[\w\.\"]+)\s*",
            re.IGNORECASE | re.DOTALL,
        )
        masking_policy_matches = masking_policy_pattern.findall(statement)
        dependencies.extend(masking_policy_matches)

        dependencies.extend(
            DependencyParser._parse_object_dependencies_views_functions_policies_static(
                statement, function_name_list
            )
        )
    
        return dependencies

    @staticmethod
    def _parse_object_dependencies_views_functions_policies_static(
        statement: str, function_name_list: List[str]
    ) -> List[str]:
        """
            Parses view and function dependencies from a given statement
            Will check for FROM statements, explicit and implicit joins
        Args:
            statement: str - the whole sql statement (multiple lines)
            function_name_list: List[str] = [] - list of function names which might be a dependency for this statement
        """
        statement_without_comment = string_util.remove_comment(statement)
        statement_clean = string_util.remove_text_in_quotes(statement_without_comment)

        explicit_pattern = re.compile(
            r"(?:\s|^)(?:FROM|JOIN)\s+(?:\(\s*?(?!SELECT))?@?([\"\w\-\_\.]+)",
            re.IGNORECASE | re.DOTALL,
        )
        implicit_pattern1 = re.compile(
            r"(?:\s|^)FROM\s+?(?P<listOfObjects>[\"\w\-\_\.]+(?:\s*?AS)?\s*?\w*?(\s*?,\s*?([\"\w\-\_\.]+)(\s*?AS)?\s*\w*\s?)+)",
            re.IGNORECASE | re.DOTALL,
        )
        implicit_pattern2 = re.compile(
            r"(?:^|,)\s*?(?P<objectname>[\"\w\-\_\.]+)", re.IGNORECASE | re.DOTALL
        )

        object_dependencies = []
        object_dependencies.extend(explicit_pattern.findall(statement_clean))

        implicit_objects = implicit_pattern1.finditer(statement_clean)
        for implicit_object in implicit_objects:
            object_dependencies.extend(
                implicit_pattern2.findall(implicit_object.groupdict()["listOfObjects"])
            )

        # find file formats
        fileformat_pattern = re.compile(
            r"file_format\s*=>\s*'?([\"\w\-\_\.]+)'?",
            re.IGNORECASE | re.DOTALL,
        )
        object_dependencies.extend(
            fileformat_pattern.findall(statement_without_comment)
        )

        # find streams
        stream_pattern = re.compile(
            r"stream\s*=>\s*'([\"\w\-\_\.]+)'",
            re.IGNORECASE | re.DOTALL,
        )
        object_dependencies.extend(stream_pattern.findall(statement_without_comment))

        if function_name_list:
            function_names = f"{'|'.join([re.escape(n) for n in function_name_list])}"
            function_pattern = re.compile(
                r"\b(?P<functionName>{function_names})\s*\(".format(
                    function_names=function_names
                ),
                re.IGNORECASE | re.DOTALL,
            )
            object_dependencies.extend(
                function_pattern.findall(statement_clean.replace('"', ""))
            )  # replace double quotes to match functions referenced with double quotes

        # remove duplicates
        object_dependencies = list(
            dict.fromkeys(object_dependencies)
        )  # remove duplicates

        # matches to be skipped:
        # finds constructs like "WHERE EXTRACT(YEAR FROM DAT_BESTAETIGUNG) >= YEAR(CURRENT_TIMESTAMP)"
        extract_function_pattern = re.compile(
            r"(?<=EXTRACT)\s?\([\"\w\-\_\.\s]*(?:FROM|JOIN)\s+?([\"\w\-\_\.]+)",
            re.IGNORECASE | re.DOTALL,
        )
        extract_matches = set(extract_function_pattern.findall(statement_clean))
        # finds references to INFORMATION_SCHEMA
        system_objects = re.compile(r"information_schema", re.IGNORECASE)
        # function to remove matches to be skipped
        false_dependencies = lambda obj: (
            system_objects.match(obj) or (obj in extract_matches)
        )

        object_dependencies = [
            obj for obj in object_dependencies if not false_dependencies(obj)
        ]

        return object_dependencies

    @staticmethod
    def _parse_object_dependencies_tables(statement: str) -> List[str]:
        """
            Parses table dependencies from a given statement
            Will check for foreign keys and sequences and policies
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        statement = string_util.remove_comment(statement)
        statement = string_util.remove_text_in_quotes(statement)

        dependencies = []

        sequence_pattern = re.compile(
            r"\sDEFAULT\s+?([\"\w\-\_\.]+)\.NEXTVAL", re.IGNORECASE | re.DOTALL
        )
        sequence_matches = sequence_pattern.findall(statement)
        dependencies.extend(sequence_matches)

        fk_pattern = re.compile(
            r"(?:[\s\)])REFERENCES\s+(?P<referenced_table>[\w\.\"]+)\s*\([\w,\" ]+\)",
            re.IGNORECASE | re.DOTALL,
        )
        fk_matches = fk_pattern.findall(statement)
        dependencies.extend(fk_matches)

        masking_policy_pattern = re.compile(
            r"(?:[\s\)])MASKING\s+POLICY\s+(?P<referenced_policy>[\w\.\"]+)\s*",
            re.IGNORECASE | re.DOTALL,
        )
        masking_policy_matches = masking_policy_pattern.findall(statement)
        dependencies.extend(masking_policy_matches)

        row_access_policy_pattern = re.compile(
            r"(?:[\s\)])ROW\s+ACCESS\s+POLICY\s+(?P<referenced_policy>[\w\.\"]+)\s+ON\s*",
            re.IGNORECASE | re.DOTALL,
        )
        row_access_policy_matches = row_access_policy_pattern.findall(statement)
        dependencies.extend(row_access_policy_matches)

        return list(set(dependencies))

    @staticmethod
    def _parse_object_dependencies_externaltables(statement: str) -> List[str]:
        """
            Parses external table dependencies from a given statement
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        statement = string_util.remove_comment(statement)
        statement = string_util.remove_text_in_quotes(statement)

        dependencies = []

        fileformat_pattern = re.compile(
            r"\b(?:FILE_FORMAT|FORMAT_NAME)\s*=\s*\'?(?P<fileformat>[\w\.\"]+)",
            re.IGNORECASE | re.DOTALL,
        )
        s = fileformat_pattern.search(statement)
        if s:
            dependencies.append(s.group("fileformat"))

        stage_pattern = re.compile(
            r"\bLOCATION\s*=\s*\@?(?P<stage>[\w\.\"]+)",
            re.IGNORECASE | re.DOTALL,
        )
        s = stage_pattern.search(statement)
        if s:
            dependencies.append(s.group("stage"))

        return list(set(dependencies))

    @staticmethod
    def _parse_object_dependencies_streams(statement: str) -> List[str]:
        """
        Parses dependencies for streams (tables, views, stages, directories, external tables, dynamic tables).
        """
        statement = string_util.remove_comment(statement)

        dependencies = []

        # Table/View/External Table/Dynamic Table dependencies
        object_types = "table|view|external table|dynamic table"
        object_pattern = re.compile(
            rf"on\s+({object_types})\s+((?:\"[^\"]+\"\.)?\"?[a-zA-Z0-9_]+\"?(?:\.[a-zA-Z0-9_]+|(?:\"\.[^\"]+\")?)*)",
            re.IGNORECASE,
        )
        for match in object_pattern.finditer(statement):
            dependencies.append(match.group(2).strip())

        # Stage dependencies
        stage_pattern = re.compile(
            r"on\s+stage\s+(@?(?:\"[^\"]+\"\.)?\"?[a-zA-Z0-9_]+\"?(?:\.[a-zA-Z0-9_]+|(?:\"\.[^\"]+\")?)*)",
            re.IGNORECASE,
        )
        for match in stage_pattern.finditer(statement):
            dependencies.append(match.group(1).lstrip("@").strip())

        # Directory dependencies (with @)
        directory_pattern = re.compile(
            r"on\s+directory\s*\(\s*@((?:\"[^\"]+\"\.)?\"?[a-zA-Z0-9_]+\"?(?:\.[a-zA-Z0-9_]+|(?:\"\.[^\"]+\")?)*)\s*\)",
            re.IGNORECASE,
        )
        for match in directory_pattern.finditer(statement):
            dependencies.append(match.group(1).strip())

        return dependencies

    @staticmethod
    def _parse_object_dependencies_tasks(statement: str) -> List[str]:
        """
            Parses table or view dependencies from a given TASK definition
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        statement = string_util.remove_comment(statement)
        statement = string_util.remove_text_in_quotes(statement)

        pattern = re.compile(
            r"\sAFTER\s+(?P<previous_tasks>(([\w\.\"]+\s*,\s*)*[\w\.\"]+))",
            re.IGNORECASE | re.DOTALL,
        )
        s = pattern.search(statement)
        if s:
            previous_tasks = s.group("previous_tasks")
            previous_tasks_list = [t.strip() for t in previous_tasks.split(",")]
            return previous_tasks_list
        else:
            return []

    @staticmethod
    def _parse_object_dependencies_stages(statement: str) -> List[str]:
        """
            Parses table or view dependencies from a given STAGE definition
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        statement = string_util.remove_comment(statement)

        pattern = re.compile(
            r"\b(?:FILE_FORMAT|FORMAT_NAME)\s*=\s*\'?(?P<fileformat>[\w\.\"]+)",
            re.IGNORECASE | re.DOTALL,
        )
        s = pattern.search(statement)
        if s:
            return [s.group("fileformat")]
        else:
            return []

    @staticmethod
    def _parse_object_dependencies_pipes(statement: str) -> List[str]:
        """
            Parses table or view dependencies from a given PIPE definition
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        dependencies = []
        statement = string_util.remove_comment(statement)

        pattern_stages = re.compile(
            r"\sFROM\s+\@(?P<stage>[\w\.\"]+)", re.IGNORECASE | re.DOTALL
        )
        s = pattern_stages.search(statement)
        if s:
            dependencies.append(s.group("stage"))

        pattern_copy_into = re.compile(
            r"\bCOPY\s+INTO\s+(?P<table>[\w\.\"]+)", re.IGNORECASE | re.DOTALL
        )
        s = pattern_copy_into.search(statement)
        if s:
            dependencies.append(s.group("table"))

        pattern_fileformat = re.compile(
            r"\b(?:FILE_FORMAT|FORMAT_NAME)\s*=\s*\'?(?P<fileformat>[\w\.\"]+)",
            re.IGNORECASE | re.DOTALL,
        )
        s = pattern_fileformat.search(statement)
        if s:
            dependencies.append(s.group("fileformat"))

        return dependencies

    def _parse_object_dependencies_dynamictables(self, statement: str) -> List[str]:
        """
        Parses dynamic table dependencies from a given statement
        Will check for FROM statements, explicit and implicit joins and functions
        """
        return DependencyParser._parse_object_dependencies_dynamictables_static(
                statement, self.function_name_list
            )


    def _parse_object_dependencies_hybridtables(self, statement: str) -> List[str]:
        """
            Parses table or view dependencies from a given HYBRID TABLE definition
        Args:
            statement: str - the whole sql statement (multiple lines)
        """
        return DependencyParser._parse_object_dependencies_tables(statement)


    @staticmethod
    def _parse_object_dependencies_dynamictables_static(
        statement: str, function_name_list: List[str] = None
    ) -> List[str]:
        """
            Parses dynamic table dependencies from a given statement
            Will check for FROM statements, explicit and implicit joins, and policies
        Args:
            statement: str - the whole sql statement (multiple lines)
            function_name_list: List[str] - (optional) list of function names which might be a dependency for this statement
        """
        if function_name_list is None:
            function_name_list = []

        dependencies = []
        
        # TODO is this the best way of getting the dependencies? -> use column INPUTS from DYNAMIC_TABLE_GRAPH_HISTORY instead?
        dependencies.extend(
            DependencyParser._parse_object_dependencies_views_static(
                statement, function_name_list
            )
        )
    
        return dependencies


from typing import List

from acedeploy.core.model_object_action_entities import DbObjectAction
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
from acedeploy.services.solution_service import SolutionClient


def generate_actions_from_git_diff(
    solution_client: SolutionClient,
    always_create_tags: bool = True
) -> List[DbObjectAction]:
    """
        Generate list of db object actions for deployment of git diff to meta db
    Returns:
        List[DbObjectAction] - List of db object actions
    Notes:
        When view was added, dependend objects of the views need to be added
    """
    object_action_list = []

    # always add all schemas; always add all tags (if config parameter is set); add all objects with git change type A, M, R
    for obj in solution_client.all_objects:
        if (
            obj.object_type == DbObjectType.SCHEMA
        ) or (
            obj.object_type == DbObjectType.TAG and always_create_tags
        ) or (
            obj.git_change_type in ("A", "M", "R")
        ):
            action = DbObjectAction.factory_from_solution_object(obj, DbActionType.ADD)
            object_action_list.append(action)

    return object_action_list


def generate_actions_from_solution(solution_client) -> List[DbObjectAction]:
    """
        Generates list of db actions for deployment based on solution items
    Returns:
        List[DbObjectAction] - List of db object actions
    """
    object_action_list = []

    for obj in solution_client.all_objects:
        action = DbObjectAction.factory_from_solution_object(obj, DbActionType.ADD)
        object_action_list.append(action)

    return object_action_list



import git


class GitFile(object):
    """
    New object to store git information on a file
    """

    def __init__(
        self, file_name: str, git_change_type: git.DiffIndex.change_type
    ) -> None:
        self.file_name = str(file_name)
        self.change_type = git_change_type.upper()
